# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import socket
from typing import Any
from collections import namedtuple
from collections.abc import Sequence

from xpra.util.system import get_linux_distribution, get_generic_os_name, do_get_generic_os_name
from xpra.util.io import load_binary_file
from xpra.platform.paths import get_icon_filename
from xpra.log import Logger

log = Logger("shadow")


def get_os_icons() -> Sequence[tuple[int, int, str, bytes]]:
    try:
        from PIL import Image  # pylint: disable=import-outside-toplevel
    except ImportError:
        return ()
    filename = (get_generic_os_name() or "").lower() + ".png"
    icon_name = get_icon_filename(filename)
    if not icon_name:
        log(f"get_os_icons() no icon matching {filename!r}")
        return ()
    with log.trap_error(f"Error: failed to load window icon {icon_name!r}"):
        img = Image.open(icon_name)
        log(f"Image({icon_name})={img}")
        if img:
            icon_data = load_binary_file(icon_name)
            if not icon_data:
                log(f"icon {icon_name} not found")
                return ()
            w, h = img.size
            img.close()
            icon = (w, h, "png", icon_data)
            icons = (icon,)
            return icons
    return ()


class RootWindowModel:
    __slots__ = (
        "window", "title", "geometry", "capture",
        "property_names", "dynamic_property_names", "internal_property_names",
        "signal_listeners",
    )

    def __init__(self, root_window, capture=None, title: str = "", geometry=None):
        self.window = root_window
        self.title: str = title
        self.geometry = geometry
        self.capture = capture
        self.property_names: list[str] = [
            "title", "class-instance",
            "client-machine", "window-type",
            "size-hints", "icons", "shadow",
            "depth",
        ]
        self.dynamic_property_names: list[str] = []
        self.internal_property_names: list[str] = ["content-type"]
        self.signal_listeners: dict[str, list[tuple]] = {}

    def __repr__(self):
        return f"RootWindowModel({self.capture} : {str(self.geometry):24})"

    def get_info(self) -> dict[str, Any]:
        info = {}
        c = self.capture
        if c:
            info["capture"] = c.get_info()
        return info

    def take_screenshot(self):
        return self.capture.take_screenshot()

    def get_image(self, x: int, y: int, width: int, height: int):
        ox, oy = self.geometry[:2]
        image = self.capture.get_image(ox + x, oy + y, width, height)
        if image and (ox > 0 or oy > 0):
            # adjust x and y of where the image is displayed on the client (target_x and target_y)
            # not where the image lives within the current buffer (x and y)
            image.set_target_x(x)
            image.set_target_y(y)
        return image

    def unmanage(self, exiting=False) -> None:
        """ subclasses may want to perform cleanup here """

    def suspend(self) -> None:
        """ subclasses may want to suspend pixel capture """

    def is_managed(self) -> bool:
        return True

    def is_tray(self) -> bool:
        return False

    def is_OR(self) -> bool:
        return False

    def has_alpha(self) -> bool:
        return False

    def uses_xshm(self) -> bool:
        return False

    def is_shadow(self) -> bool:
        return True

    def get_default_window_icon(self, _size: int = 48):
        return None

    def acknowledge_changes(self) -> None:
        """ only window models that use the X11 Damage extension use this method """

    def get_dimensions(self):
        # used by get_window_info only
        return self.geometry[2:4]

    def get_geometry(self):
        return self.geometry

    def get_property_names(self) -> list[str]:
        return self.property_names

    def get_dynamic_property_names(self) -> list[str]:
        return self.dynamic_property_names

    def get_internal_property_names(self) -> list[str]:
        return self.internal_property_names

    def get_property(self, prop: str):
        # subclasses can define properties as attributes:
        attr_name = prop.replace("-", "_")
        if hasattr(self, attr_name):
            return getattr(self, attr_name)
        # otherwise fallback to default behaviour:
        if prop == "title":
            return self.title
        if prop == "client-machine":
            return socket.gethostname()
        if prop == "window-type":
            return ["NORMAL"]
        if prop == "fullscreen":
            return False
        if prop == "shadow":
            return True
        if prop == "depth":
            return 24
        if prop == "scaling":
            return None
        if prop == "opacity":
            return None
        if prop == "size-hints":
            size = self.get_dimensions()
            return {
                "maximum-size": size,
                "minimum-size": size,
                "base-size": size,
            }
        if prop == "class-instance":
            osn = do_get_generic_os_name()
            if osn == "Linux":
                try:
                    osn += "-" + get_linux_distribution()[0].replace(" ", "-")
                except Exception:  # pragma: no cover
                    pass
            return f"xpra-{osn.lower()}", f"Xpra {osn.replace('-', ' ')}"
        if prop == "icons":
            return get_os_icons()
        if prop == "content-type":
            return "desktop"
        raise AttributeError(f"invalid property {prop!r}")

    def get(self, name: str, default_value=None):
        try:
            return self.get_property(name)
        except AttributeError as e:
            log("get(%s, %s) %s on %s", name, default_value, e, self)
            return default_value

    def notify(self, prop: str) -> None:
        listeners = self.signal_listeners.get(prop)
        if listeners is None:
            log.warn(f"Warning: ignoring notify for {prop!r}")
            return
        PSpec = namedtuple("PSpec", ("name", ))
        pspec = PSpec(name=prop)
        for listener, *args in listeners:
            with log.trap_error(f"Error on {prop!r} signal listener {listener}"):
                listener(self, pspec, *args)

    def managed_connect(self, signal: str, *args):  # pragma: no cover
        self.connect(signal, *args)

    def connect(self, signal: str, *args):  # pragma: no cover
        prop = signal.split(":")[-1]  # notify::geometry
        if prop not in self.dynamic_property_names:
            log.warn(f"Warning: ignoring signal connect request: {args}")
            return
        self.signal_listeners.setdefault(prop, []).append(args)

    def disconnect(self, *args):  # pragma: no cover
        log.warn(f"Warning: ignoring signal disconnect request: {args}")



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any, Final

from xpra.util.env import envbool
from xpra.os_util import gi_import
from xpra.common import MAX_WINDOW_SIZE
from xpra.gtk.error import xsync, xswallow, xlog
from xpra.gtk.gobject import no_arg_signal, one_arg_signal
from xpra.gtk.util import get_default_root_window
from xpra.x11.common import Unmanageable
from xpra.x11.gtk.native_window import GDKX11Window
from xpra.x11.gtk.selection import ManagerSelection
from xpra.x11.gtk.prop import prop_set, prop_get, prop_del, raw_prop_set, prop_encode
from xpra.x11.gtk.world_window import WorldWindow, destroy_world_window
from xpra.x11.gtk.bindings import add_event_receiver, add_fallback_receiver, remove_fallback_receiver
from xpra.x11.models.window import WindowModel, configure_bits
from xpra.x11.window_info import window_name, window_info
from xpra.x11.bindings.window import constants, X11WindowBindings
from xpra.x11.bindings.keyboard import X11KeyboardBindings
from xpra.log import Logger

GObject = gi_import("GObject")
Gdk = gi_import("Gdk")

log = Logger("x11", "window")

X11Window = X11WindowBindings()
X11Keyboard = X11KeyboardBindings()

focuslog = Logger("x11", "window", "focus")
screenlog = Logger("x11", "window", "screen")
framelog = Logger("x11", "window", "frame")

CWX: Final[int] = constants["CWX"]
CWY: Final[int] = constants["CWY"]
CWWidth: Final[int] = constants["CWWidth"]
CWHeight: Final[int] = constants["CWHeight"]

NotifyPointerRoot: Final[int] = constants["NotifyPointerRoot"]
NotifyDetailNone: Final[int] = constants["NotifyDetailNone"]

LOG_MANAGE_FAILURES = envbool("XPRA_LOG_MANAGE_FAILURES", False)

NO_NET_SUPPORTED = os.environ.get("XPRA_NO_NET_SUPPORTED", "").split(",")

DEFAULT_NET_SUPPORTED = [
    "_NET_SUPPORTED",  # a bit redundant, perhaps...
    "_NET_SUPPORTING_WM_CHECK",
    "_NET_WM_FULL_PLACEMENT",
    "_NET_WM_HANDLED_ICONS",
    "_NET_CLIENT_LIST",
    "_NET_CLIENT_LIST_STACKING",
    "_NET_DESKTOP_VIEWPORT",
    "_NET_DESKTOP_GEOMETRY",
    "_NET_NUMBER_OF_DESKTOPS",
    "_NET_DESKTOP_NAMES",
    "_NET_WORKAREA",
    "_NET_ACTIVE_WINDOW",
    "_NET_CURRENT_DESKTOP",
    "_NET_SHOWING_DESKTOP",

    "WM_NAME", "_NET_WM_NAME",
    "WM_ICON_NAME", "_NET_WM_ICON_NAME",
    "WM_ICON_SIZE",
    "WM_CLASS",
    "WM_PROTOCOLS",
    "_NET_WM_PID",
    "WM_CLIENT_MACHINE",
    "WM_STATE",

    "_NET_WM_FULLSCREEN_MONITORS",

    "_NET_WM_ALLOWED_ACTIONS",
    "_NET_WM_ACTION_CLOSE",
    "_NET_WM_ACTION_FULLSCREEN",

    # We don't actually use _NET_WM_USER_TIME at all (yet), but it is
    # important to say we support the _NET_WM_USER_TIME_WINDOW property,
    # because this tells applications that they do not need to constantly
    # ping any pagers etc. that might be running -- see EWMH for details.
    # (Though it's not clear that any applications actually take advantage
    # of this yet.)
    "_NET_WM_USER_TIME",
    "_NET_WM_USER_TIME_WINDOW",
    # Not fully:
    "WM_HINTS",
    "WM_NORMAL_HINTS",
    "WM_TRANSIENT_FOR",
    "_NET_WM_STRUT",
    "_NET_WM_STRUT_PARTIAL"
    "_NET_WM_ICON",

    "_NET_CLOSE_WINDOW",

    # These aren't supported in any particularly meaningful way, but hey.
    "_NET_WM_WINDOW_TYPE",
    "_NET_WM_WINDOW_TYPE_NORMAL",
    "_NET_WM_WINDOW_TYPE_DESKTOP",
    "_NET_WM_WINDOW_TYPE_DOCK",
    "_NET_WM_WINDOW_TYPE_TOOLBAR",
    "_NET_WM_WINDOW_TYPE_MENU",
    "_NET_WM_WINDOW_TYPE_UTILITY",
    "_NET_WM_WINDOW_TYPE_SPLASH",
    "_NET_WM_WINDOW_TYPE_DIALOG",
    "_NET_WM_WINDOW_TYPE_DROPDOWN_MENU",
    "_NET_WM_WINDOW_TYPE_POPUP_MENU",
    "_NET_WM_WINDOW_TYPE_TOOLTIP",
    "_NET_WM_WINDOW_TYPE_NOTIFICATION",
    "_NET_WM_WINDOW_TYPE_COMBO",
    # "_NET_WM_WINDOW_TYPE_DND",

    "_NET_WM_STATE",
    "_NET_WM_STATE_DEMANDS_ATTENTION",
    "_NET_WM_STATE_MODAL",
    # More states to support:
    "_NET_WM_STATE_STICKY",
    "_NET_WM_STATE_MAXIMIZED_VERT",
    "_NET_WM_STATE_MAXIMIZED_HORZ",
    "_NET_WM_STATE_SHADED",
    "_NET_WM_STATE_SKIP_TASKBAR",
    "_NET_WM_STATE_SKIP_PAGER",
    "_NET_WM_STATE_HIDDEN",
    "_NET_WM_STATE_FULLSCREEN",
    "_NET_WM_STATE_ABOVE",
    "_NET_WM_STATE_BELOW",
    "_NET_WM_STATE_FOCUSED",

    "_NET_WM_DESKTOP",

    "_NET_WM_MOVERESIZE",
    "_NET_MOVERESIZE_WINDOW",

    "_MOTIF_WM_HINTS",
    "_MOTIF_WM_INFO",

    "_NET_REQUEST_FRAME_EXTENTS",
    "_NET_RESTACK_WINDOW",

    "_NET_WM_OPAQUE_REGION",
]
FRAME_EXTENTS = envbool("XPRA_FRAME_EXTENTS", True)
if FRAME_EXTENTS:
    DEFAULT_NET_SUPPORTED.append("_NET_FRAME_EXTENTS")

NET_SUPPORTED = [x for x in DEFAULT_NET_SUPPORTED if x not in NO_NET_SUPPORTED]

DEFAULT_SIZE_CONSTRAINTS = (0, 0, MAX_WINDOW_SIZE, MAX_WINDOW_SIZE)


class Wm(GObject.GObject):
    __gproperties__ = {
        "windows": (GObject.TYPE_PYOBJECT,
                    "Set of managed windows (as WindowModels)", "",
                    GObject.ParamFlags.READABLE),
        "toplevel": (GObject.TYPE_PYOBJECT,
                     "Toplevel container widget for the display", "",
                     GObject.ParamFlags.READABLE),
    }
    __gsignals__ = {
        # Public use:
        # A new window has shown up:
        "new-window": one_arg_signal,
        "show-desktop": one_arg_signal,
        # You can emit this to cause the WM to quit, or the WM may
        # spontaneously raise it if another WM takes over the display.  By
        # default, unmanages all windows:
        "quit": no_arg_signal,

        # Mostly intended for internal use:
        "x11-child-map-request-event": one_arg_signal,
        "x11-child-configure-request-event": one_arg_signal,
        "x11-focus-in-event": one_arg_signal,
        "x11-focus-out-event": one_arg_signal,
        "x11-client-message-event": one_arg_signal,
        "x11-xkb-event": one_arg_signal,
    }

    def __init__(self, replace_other_wm: bool, wm_name: str):
        super().__init__()

        self._wm_name = wm_name
        self._ewmh_window = None

        self._windows: dict[int, Any] = {}
        # EWMH says we have to know the order of our windows oldest to
        # youngest...
        self._windows_in_order = []

        # Become the Official Window Manager of this year's display:
        self._wm_selection = ManagerSelection("WM_S0")
        self._cm_wm_selection = ManagerSelection("_NET_WM_CM_S0")
        self._wm_selection.connect("selection-lost", self._lost_wm_selection)
        self._cm_wm_selection.connect("selection-lost", self._lost_wm_selection)
        # May throw AlreadyOwned:
        if replace_other_wm:
            mode = self._wm_selection.FORCE
        else:
            mode = self._wm_selection.IF_UNOWNED
        self._wm_selection.acquire(mode)
        self._cm_wm_selection.acquire(mode)

        # Set up the necessary EWMH properties on the root window.
        self._setup_ewmh_window()
        # Start with just one desktop:
        self.set_desktop_list(("Main",))
        self.set_current_desktop(0)
        # Start with the full display as workarea:
        root = get_default_root_window()
        root_w, root_h = root.get_geometry()[2:4]
        self.root_set("_NET_SUPPORTED", ["atom"], NET_SUPPORTED)
        self.set_workarea(0, 0, root_w, root_h)
        self.set_desktop_geometry(root_w, root_h)
        self.root_set("_NET_DESKTOP_VIEWPORT", ["u32"], [0, 0])

        self.size_constraints = DEFAULT_SIZE_CONSTRAINTS

        # Load up our full-screen widget
        self._world_window = WorldWindow()
        self.notify("toplevel")
        self._world_window.show_all()
        wxid = self._world_window.get_window().get_xid()

        # Okay, ready to select for SubstructureRedirect and then load in all
        # the existing clients.
        rxid = root.get_xid()
        add_event_receiver(rxid, self)
        add_fallback_receiver("x11-client-message-event", self)
        # when reparenting, the events may get sent
        # to a window that is already destroyed,
        # and we don't want to miss those events, so:
        add_fallback_receiver("x11-child-map-request-event", self)
        X11Window.substructureRedirect(rxid)

        children = X11Window.get_children(rxid)
        log(f"root window children: {children}")
        for xid in children:
            # ignore windows we have created ourselves (ie: the world window),
            # unmapped or `OR` windows:
            if xid == wxid:
                continue
            if X11Window.is_override_redirect(xid):
                log("skipping %s", window_info(xid))
                continue
            if not X11Window.is_mapped(xid):
                log("skipping %s", window_info(xid))
                continue
            log(f"Wm managing pre-existing child window {xid:x}")
            self._manage_client(xid)

        # Also watch for focus change events on the root window
        X11Window.selectFocusChange(rxid)
        X11Keyboard.selectBellNotification(True)
        X11Window.setRootIconSizes(64, 64)

        # FIXME:
        # Need viewport abstraction for _NET_CURRENT_DESKTOP...
        # Tray's need to provide info for _NET_ACTIVE_WINDOW and _NET_WORKAREA
        # (and notifications for both)

    @staticmethod
    def root_set(*args) -> None:
        prop_set(X11Window.get_root_xid(), *args)

    @staticmethod
    def root_get(*args, **kwargs):
        return prop_get(X11Window.get_root_xid(), *args, **kwargs)

    def set_workarea(self, x: int, y: int, width: int, height: int) -> None:
        v = [x, y, width, height]
        screenlog("_NET_WORKAREA=%s", v)
        self.root_set("_NET_WORKAREA", ["u32"], v)

    def set_desktop_geometry(self, width: int, height: int) -> None:
        v = [width, height]
        screenlog("_NET_DESKTOP_GEOMETRY=%s", v)
        self.root_set("_NET_DESKTOP_GEOMETRY", ["u32"], v)
        # update all the windows:
        for model in self._windows.values():
            model.update_desktop_geometry(width, height)

    def set_size_constraints(self, minw: int = 0, minh: int = 0, maxw: int = MAX_WINDOW_SIZE,
                             maxh: int = MAX_WINDOW_SIZE) -> None:
        log("set_size_constraints%s", (minw, minh, maxw, maxh))
        self.size_constraints = minw, minh, maxw, maxh
        # update all the windows:
        for model in self._windows.values():
            model.update_size_constraints(minw, minh, maxw, maxh)

    def set_default_frame_extents(self, v):
        framelog("set_default_frame_extents(%s)", v)
        if not v or len(v) != 4:
            v = (0, 0, 0, 0)
        self.root_set("DEFAULT_NET_FRAME_EXTENTS", ["u32"], v)
        # update the models that are using the global default value:
        for win in self._windows.values():
            if win.is_OR() or win.is_tray():
                continue
            cur = win.get_property("frame")
            if cur is None:
                win._handle_frame_changed()

    def do_get_property(self, pspec):
        if pspec.name == "windows":
            return frozenset(self._windows.values())
        if pspec.name == "toplevel":
            return self._world_window
        assert False

    # This is in some sense the key entry point to the entire WM program.  We
    # have detected a new client window, and start managing it:
    def _manage_client(self, xid: int):
        if xid in self._windows:
            # already managed
            return
        try:
            with xsync:
                log("_manage_client(%x)", xid)
                desktop_geometry = self.root_get("_NET_DESKTOP_GEOMETRY", ["u32"], True, False)
                root = get_default_root_window()
                win = WindowModel(root.get_xid(), xid, desktop_geometry, self.size_constraints)
        except Exception as e:
            if LOG_MANAGE_FAILURES or not isinstance(e, Unmanageable):
                log_fn = log.warn
            else:
                log_fn = log.debug
            log_fn("Warning: failed to manage client window %#x:", xid)
            log_fn(" %s", e)
            log_fn("", exc_info=True)
            with xswallow:
                log_fn(" window name: %s", window_name(xid))
                log_fn(" window info: %s", window_info(xid))
        else:
            win.managed_connect("unmanaged", self._handle_client_unmanaged, xid)
            self._windows[xid] = win
            self._windows_in_order.append(xid)
            self.notify("windows")
            self._update_window_list()
            self.emit("new-window", win)

    def _handle_client_unmanaged(self, model, wm_exiting, xid: int) -> None:
        log(f"_handle_client_unmanaged({model}, {wm_exiting}, {xid:x})")
        if xid not in self._windows:
            log.error(f"Error: gdk window {xid} not found in {self._windows}")
            return
        del self._windows[xid]
        self._windows_in_order.remove(xid)
        self._update_window_list()
        self.notify("windows")

    def _update_window_list(self, *_args) -> None:
        # Ignore errors because not all the windows may still exist; if so,
        # then it's okay to leave the lists out of date for a moment, because
        # in a moment we'll get a signal telling us about the window that
        # doesn't exist anymore, will remove it from the list, and then call
        # _update_window_list again.
        dtype, dformat, window_xids = prop_encode(["u32"], tuple(self._windows_in_order))
        log("prop_encode(%s)=%s", self._windows_in_order, (dtype, dformat, window_xids))
        xid = X11Window.get_root_xid()
        with xlog:
            for prop in ("_NET_CLIENT_LIST", "_NET_CLIENT_LIST_STACKING"):
                raw_prop_set(xid, prop, "WINDOW", dformat, window_xids)

    def do_x11_client_message_event(self, event) -> None:
        # FIXME
        # Need to listen for:
        #   _NET_ACTIVE_WINDOW
        #   _NET_CURRENT_DESKTOP
        #   _NET_WM_PING responses
        # and maybe:
        #   _NET_WM_STATE
        log("do_x11_client_message_event(%s)", event)
        if event.message_type == "_NET_SHOWING_DESKTOP":
            show = bool(event.data[0])
            self.emit("show-desktop", show)
        elif event.message_type == "_NET_REQUEST_FRAME_EXTENTS" and FRAME_EXTENTS:
            # if we're here, that means the window model does not exist
            # (or it would have processed the event)
            # so this must be an unmapped window
            frame = None
            with xswallow:
                xid = event.window
                if not X11Window.is_override_redirect(xid):
                    # use the global default:
                    frame = self.root_get("DEFAULT_NET_FRAME_EXTENTS", ["u32"], ignore_errors=True)
                if not frame:
                    # fallback:
                    frame = (0, 0, 0, 0)
                framelog("_NET_REQUEST_FRAME_EXTENTS: setting _NET_FRAME_EXTENTS=%s on %#x", frame, xid)
                prop_set(event.window, "_NET_FRAME_EXTENTS", ["u32"], frame)

    def _lost_wm_selection(self, selection) -> None:
        log.info("Lost WM selection %s, exiting", selection)
        self.emit("quit")

    def do_quit(self) -> None:
        self.cleanup()

    def cleanup(self) -> None:
        remove_fallback_receiver("x11-client-message-event", self)
        remove_fallback_receiver("x11-child-map-request-event", self)
        for win in tuple(self._windows.values()):
            win.unmanage(True)
        xid = self._ewmh_window.get_xid()
        with xswallow:
            prop_del(xid, "_NET_SUPPORTING_WM_CHECK")
            prop_del(xid, "_NET_WM_NAME")
        destroy_world_window()

    def do_x11_child_map_request_event(self, event) -> None:
        log("Found a potential client")
        self._manage_client(event.window)

    def do_x11_child_configure_request_event(self, event) -> None:
        # The point of this method is to handle configure requests on
        # withdrawn windows.  We simply allow them to move/resize any way they
        # want.  This is harmless because the window isn't visible anyway (and
        # apps can create unmapped windows with whatever coordinates they want
        # anyway, no harm in letting them move existing ones around), and it
        # means that when the window actually gets mapped, we have more
        # accurate info on what the app is actually requesting.
        xid = event.window
        if not xid:
            return
        model = self._windows.get(xid)
        if model:
            # the window has been reparented already,
            # but we're getting the configure request event on the root window
            # forward it to the model
            log("do_x11_child_configure_request_event(%s) value_mask=%s, forwarding to %s",
                event, configure_bits(event.value_mask), model)
            model.do_x11_child_configure_request_event(event)
            return
        log("do_x11_child_configure_request_event(%s) value_mask=%s, reconfigure on withdrawn window",
            event, configure_bits(event.value_mask))
        with xswallow:
            geom = X11Window.getGeometry(xid)
            if not geom:
                log(f"failed to get geometry for window {xid:x} - skipping configure request")
                return
            x, y, w, h = geom[:4]
            if event.value_mask & CWX:
                x = event.x
            if event.value_mask & CWY:
                y = event.y
            if event.value_mask & CWWidth:
                w = event.width
            if event.value_mask & CWHeight:
                h = event.height
            if event.value_mask & (CWX | CWY | CWWidth | CWHeight):
                log("updated window geometry for window %#x from %s to %s",
                    xid, geom[:4], (x, y, w, h))
            X11Window.configure(xid, x, y, w, h, event.value_mask)
            X11Window.sendConfigureNotify(xid)

    def do_x11_focus_in_event(self, event) -> None:
        # The purpose of this function is to detect when the focus mode has
        # gone to PointerRoot or None, so that it can be given back to
        # something real.  This is easy to detect -- a FocusIn event with
        # detail PointerRoot or None is generated on the root window.
        focuslog("wm.do_x11_focus_in_event(%s)", event)
        if event.detail in (NotifyPointerRoot, NotifyDetailNone) and self._world_window:
            self._world_window.reset_x_focus()

    # noinspection PyMethodMayBeStatic
    def do_x11_focus_out_event(self, event) -> None:
        focuslog("wm.do_x11_focus_out_event(%s) XGetInputFocus=%s", event, X11Window.XGetInputFocus())

    def set_desktop_list(self, desktops) -> None:
        log("set_desktop_list(%s)", desktops)
        self.root_set("_NET_NUMBER_OF_DESKTOPS", "u32", len(desktops))
        self.root_set("_NET_DESKTOP_NAMES", ["utf8"], desktops)

    def set_current_desktop(self, index) -> None:
        self.root_set("_NET_CURRENT_DESKTOP", "u32", index)

    def _setup_ewmh_window(self) -> None:
        # Set up a 1x1 invisible unmapped window, with which to participate in
        # EWMH's _NET_SUPPORTING_WM_CHECK protocol.  The only important things
        # about this window are the _NET_SUPPORTING_WM_CHECK property, and
        # its title (which is supposed to be the name of the window manager).

        # NB, GDK will do strange things to this window.  We don't want to use
        # it for anything.  (In particular, it will call XSelectInput on it,
        # which is fine normally when GDK is running in a client, but since it
        # happens to be using the same connection as we, the WM, it will
        # clobber any `XSelectInput` calls that *we* might have wanted to make
        # on this window.)  Also, GDK might silently swallow all events that
        # are detected on it, anyway.
        root = get_default_root_window()
        self._ewmh_window = GDKX11Window(root, wclass=Gdk.WindowWindowClass.INPUT_ONLY, title=self._wm_name)
        xid = self._ewmh_window.get_xid()
        prop_set(xid, "_NET_SUPPORTING_WM_CHECK", "window", xid)
        self.root_set("_NET_SUPPORTING_WM_CHECK", "window", xid)
        self.root_set("_NET_WM_NAME", "utf8", self._wm_name)

    def get_net_wm_name(self) -> str:
        try:
            return prop_get(self._ewmh_window.get_xid(), "_NET_WM_NAME", "utf8",
                            ignore_errors=False, raise_xerrors=False)
        except Exception as e:
            log.error("Error querying _NET_WM_NAME")
            log.estr(e)
            return ""


GObject.type_register(Wm)



# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


import sys
from xpra.net.vsock import bind_vsocket, CID_HOST, PORT_ANY       #@UnresolvedImport

from xpra.log import Logger
log = Logger("network")


def main(args):
    cid = CID_HOST
    port = PORT_ANY
    if len(args)>0:
        cid = int(args[0])
    if len(args)>1:
        port = int(args[1])
    sock = bind_vsocket(cid=cid, port=port)
    log("sock=%s", sock)
    sock.listen(1)
    log("listening")
    while True:
        connection, client_address = sock.accept()
        log("new connection! %s", (connection, client_address))
        l = 0
        import time
        start = time.time()
        while True:
            buf = connection.recv(1024*1024)
            if buf:
                log("got data: %s" % len(buf))
                l += len(buf)
                now = time.time()
                log("speed=%iMB/s", (l/1024/1024/(now-start)))
            else:
                return 0
        connection.send("hello")

if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# This file is part of Xpra.
# Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# pylint: disable-msg=E1101

from typing import Any

from xpra.net.common import PacketType
from xpra.util.objects import typedict
from xpra.util.env import envbool
from xpra.client.base.stub_client_mixin import StubClientMixin
from xpra.log import Logger

log = Logger("cursor")

SAVE_CURSORS: bool = envbool("XPRA_SAVE_CURSORS", False)


class CursorClient(StubClientMixin):
    """
    Add cursor handling
    """
    PREFIX = "cursor"

    def __init__(self):
        self.server_cursors: bool = False
        self.client_supports_cursors: bool = False
        self.cursors_enabled: bool = False
        self.default_cursor_data = None

    def init(self, opts) -> None:
        self.client_supports_cursors = opts.cursors

    def get_info(self) -> dict[str, Any]:
        return self.get_caps()

    ######################################################################
    # hello:
    def get_caps(self) -> dict[str, Any]:
        return {
            "cursors": self.client_supports_cursors,
        }

    def parse_server_capabilities(self, c: typedict) -> bool:
        self.server_cursors = c.boolget("cursors", True)  # added in 0.5, default to True!
        self.cursors_enabled = self.server_cursors and self.client_supports_cursors
        self.default_cursor_data = c.tupleget("cursor.default", None)
        return True

    def _process_cursor(self, packet: PacketType) -> None:
        if not self.cursors_enabled:
            return
        if len(packet) == 2:
            # marker telling us to use the default cursor:
            new_cursor = packet[1]
            setdefault = False
        else:
            if len(packet) < 9:
                raise ValueError(f"invalid cursor packet: only {len(packet)} items")
            # trim packet-type:
            new_cursor = list(packet[1:])
            encoding = str(new_cursor[0])
            setdefault = encoding.startswith("default:")
            if setdefault:
                encoding = encoding.split(":")[1]
            new_cursor[0] = encoding
            if encoding == "png":
                pixels = new_cursor[8]
                if SAVE_CURSORS:
                    serial = new_cursor[7]
                    with open(f"raw-cursor-{serial:x}.png", "wb") as f:
                        f.write(pixels)
                from xpra.codecs.pillow.decoder import open_only  # pylint: disable=import-outside-toplevel
                img = open_only(pixels, ("png",))
                new_cursor[8] = img.tobytes("raw", "BGRA")
                log("used PIL to convert png cursor to raw")
                new_cursor[0] = "raw"
            elif encoding != "raw":
                log.warn(f"Warning: invalid cursor encoding: {encoding}")
                return
        if setdefault:
            log(f"setting default cursor={new_cursor!r}")
            self.default_cursor_data = new_cursor
        else:
            self.set_windows_cursor(self._id_to_window.values(), new_cursor)

    def reset_cursor(self) -> None:
        self.set_windows_cursor(self._id_to_window.values(), [])

    def set_windows_cursor(self, client_windows, new_cursor) -> None:
        raise NotImplementedError()

    def init_authenticated_packet_handlers(self) -> None:
        self.add_packets("cursor", main_thread=True)



#!/usr/bin/env python3

import gi

gi.require_version('Gtk', '3.0')  # @UndefinedVariable
gi.require_version('Gdk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk, GLib  # pylint: disable=wrong-import-position @UnresolvedImport

from xpra.x11.gtk.display_source import init_gdk_display_source
from xpra.x11.gtk.prop import prop_set
from xpra.gtk.error import xsync


def main():
    init_gdk_display_source()
    win = Gtk.Window()
    win.set_size_request(400, 100)
    win.set_title("WM_COMMAND test")
    win.show()

    def change_wmcommand():
        with xsync:
            prop_set(win.get_window(), "WM_COMMAND", "latin1", "HELLO WORLD")
            print("WM_COMMAND changed!")

    GLib.timeout_add(1000, change_wmcommand)
    Gtk.main()
    return 0


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys


def main(argv=()) -> int:
    # add `Xpra/` and `Xpra/lib` to the %PATH%
    # then call the real EXE:
    app_dir, exe_name = os.path.split(argv[0])
    exe_name = os.path.normcase(exe_name)
    if exe_name.lower().endswith(".exe"):
        exe_name = exe_name[:-4]
    lib_dir = os.path.join(app_dir, "lib")
    paths = os.environ.get("PATH", "").split(os.pathsep)
    for d in (lib_dir, app_dir):
        if d not in paths:
            paths.append(d)
    env = os.environ.copy()
    env["PATH"] = os.pathsep.join(paths)
    # try harder to find the matching executable file:
    for ext in (".exe", "-1.0.exe", ""):
        actual_exe = os.path.join(lib_dir, f"{exe_name}{ext}")
        if os.path.exists(actual_exe):
            break
    from subprocess import run
    args = [actual_exe] + list(argv[1:])
    return run(args, stdin=None, stdout=sys.stdout, stderr=sys.stderr, shell=False, env=env).returncode


if __name__ == "__main__":
    v = main(sys.argv)
    sys.exit(v)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# Low level support for the "system tray" on MS Windows
# Based on code from winswitch, itself based on "win32gui_taskbar demo"

import os
from ctypes import (
    Structure, byref, addressof, c_void_p, sizeof, POINTER,
    get_last_error, WinError, WinDLL, HRESULT,  # @UnresolvedImport
)
from ctypes.wintypes import HWND, UINT, POINT, HICON, BOOL, CHAR, WCHAR, DWORD, HMODULE, RECT
from typing import Any
from collections.abc import Callable, Sequence

from xpra.util.objects import typedict
from xpra.util.str_fn import csv, bytestostr
from xpra.util.env import envbool
from xpra.common import noop, XPRA_GUID1, XPRA_GUID2, XPRA_GUID3, XPRA_GUID4
from xpra.platform.win32 import constants as win32con
from xpra.platform.win32.icon_util import image_to_ICONINFO
from xpra.platform.win32.common import (
    GUID, WNDCLASSEX, WNDPROC,
    GetSystemMetrics,
    EnumDisplayMonitors,
    GetCursorPos,
    PostMessageA,
    CreateWindowExA, CreatePopupMenu, AppendMenu,
    LoadIconA,
    DefWindowProcA, RegisterWindowMessageA, RegisterClassExA,
    LoadImageW, DestroyIcon,
    UpdateWindow, DestroyWindow,
    PostQuitMessage,
    GetModuleHandleExA,
    GetStockObject,
)
from xpra.log import Logger

log = Logger("tray", "win32")
geomlog = Logger("tray", "geometry")

log("loading ctypes NotifyIcon functions")

TRAY_ALPHA = envbool("XPRA_TRAY_ALPHA", True)


def GetProductInfo(dwOSMajorVersion=5, dwOSMinorVersion=0, dwSpMajorVersion=0, dwSpMinorVersion=0):
    product_type = DWORD(0)
    from xpra.platform.win32.common import GetProductInfo as k32GetProductInfo
    v = k32GetProductInfo(dwOSMajorVersion, dwOSMinorVersion, dwSpMajorVersion, dwSpMinorVersion, byref(product_type))
    log("GetProductInfo(%i, %i, %i, %i)=%i product_type=%s",
        dwOSMajorVersion, dwOSMinorVersion, dwSpMajorVersion, dwSpMinorVersion, v, hex(product_type.value))
    return bool(v)


# MAX_TIP_SIZE = 128
MAX_TIP_SIZE = 64


# noinspection PyTypeChecker
def getNOTIFYICONDATAClass(char_type=CHAR, tip_size: int = MAX_TIP_SIZE):
    class _NOTIFYICONDATA(Structure):
        _fields_ = (
            ("cbSize", DWORD),
            ("hWnd", HWND),
            ("uID", UINT),
            ("uFlags", UINT),
            ("uCallbackMessage", UINT),
            ("hIcon", HICON),
            ("szTip", char_type * tip_size),
            ("dwState", DWORD),
            ("dwStateMask", DWORD),
            ("szInfo", char_type * 256),
            ("uVersion", UINT),
            ("szInfoTitle", char_type * 64),
            ("dwInfoFlags", DWORD),
            ("guidItem", GUID),
            ("hBalloonIcon", HICON),
        )

    return _NOTIFYICONDATA


NOTIFYICONDATAA = getNOTIFYICONDATAClass()
NOTIFYICONDATAW = getNOTIFYICONDATAClass(WCHAR)

shell32 = WinDLL("shell32", use_last_error=True)
Shell_NotifyIconA = shell32.Shell_NotifyIconA
Shell_NotifyIconA.restype = BOOL
Shell_NotifyIconA.argtypes = [DWORD, c_void_p]
Shell_NotifyIconW = shell32.Shell_NotifyIconW
Shell_NotifyIconW.restype = BOOL
Shell_NotifyIconW.argtypes = [DWORD, c_void_p]
Shell_NotifyIconGetRect = shell32.Shell_NotifyIconGetRect
SHSTDAPI = HRESULT


class NOTIFYICONIDENTIFIER(Structure):
    _fields_ = (
        ("cbSize", DWORD),
        ("hWnd", HWND),
        ("uID", UINT),
        ("guidItem", GUID),
    )


PNOTIFYICONIDENTIFIER = POINTER(NOTIFYICONIDENTIFIER)
Shell_NotifyIconGetRect.restype = SHSTDAPI
Shell_NotifyIconGetRect.argtypes = [PNOTIFYICONIDENTIFIER, POINTER(RECT)]

Shell_NotifyIcon = Shell_NotifyIconA
NOTIFYICONDATA = NOTIFYICONDATAA

XPRA_GUID = GUID()
XPRA_GUID.Data1 = XPRA_GUID1
XPRA_GUID.Data2 = XPRA_GUID2
XPRA_GUID.Data3 = XPRA_GUID3
XPRA_GUID.Data4 = XPRA_GUID4

FALLBACK_ICON = LoadIconA(0, win32con.IDI_APPLICATION)

# constants found in win32gui:
NIM_ADD = 0
NIM_MODIFY = 1
NIM_DELETE = 2
NIM_SETFOCUS = 3
NIM_SETVERSION = 4

NIF_MESSAGE = 1
NIF_ICON = 2
NIF_TIP = 4
NIF_STATE = 8
NIF_INFO = 16
NIF_GUID = 32
NIF_REALTIME = 64
NIF_SHOWTIP = 128

NIF_FLAGS = {
    NIF_MESSAGE: "MESSAGE",
    NIF_ICON: "ICON",
    NIF_TIP: "TIP",
    NIF_STATE: "STATE",
    NIF_INFO: "INFO",
    NIF_GUID: "GUID",
    NIF_REALTIME: "REALTIME",
    NIF_SHOWTIP: "SHOWTIP",
}

# found here:
# http://msdn.microsoft.com/en-us/library/windows/desktop/ff468877(v=vs.85).aspx
WM_XBUTTONDOWN = 0x020B
WM_XBUTTONUP = 0x020C
WM_XBUTTONDBLCLK = 0x020D

BUTTON_MAP: dict[int, list[tuple[int, int]]] = {
    win32con.WM_LBUTTONDOWN: [(1, 1)],
    win32con.WM_LBUTTONUP: [(1, 0)],
    win32con.WM_MBUTTONDOWN: [(2, 1)],
    win32con.WM_MBUTTONUP: [(2, 0)],
    win32con.WM_RBUTTONDOWN: [(3, 1)],
    win32con.WM_RBUTTONUP: [(3, 0)],
    win32con.WM_LBUTTONDBLCLK: [(1, 1), (1, 0)],
    win32con.WM_MBUTTONDBLCLK: [(2, 1), (2, 0)],
    win32con.WM_RBUTTONDBLCLK: [(3, 1), (3, 0)],
    WM_XBUTTONDOWN: [(4, 1)],
    WM_XBUTTONUP: [(4, 0)],
    WM_XBUTTONDBLCLK: [(4, 1), (4, 0)],
}


class win32NotifyIcon:
    # we register the windows event handler on the class,
    # this allows us to know which hwnd refers to which instance:
    instances: dict[int, Any] = {}

    def __init__(self, app_id: int = 0, title: str = "",
                 move_callback: Callable = noop,
                 click_callback: Callable = noop,
                 exit_callback: Callable = noop,
                 command_callback: Callable = noop,
                 iconPathName: str = ""):
        log("win32NotifyIcon: app_id=%i, title=%r", app_id, title)
        self.app_id = app_id
        self.title = title
        self.current_icon = None
        self.destroy_icon = None
        self.move_callback = move_callback
        self.click_callback = click_callback
        self.exit_callback = exit_callback
        self.command_callback = command_callback
        self.reset_function: tuple[Callable, Sequence[Any]] | None = None
        self.image_cache: dict[str, Any] = {}
        # Create the Window.
        if iconPathName:
            self.current_icon = self.LoadImage(iconPathName) or FALLBACK_ICON
        self.create_tray_window()

    def __repr__(self):
        return "win32NotifyIcon(%#x)" % self.app_id

    def create_tray_window(self) -> None:
        log("create_tray_window()")
        self.create_window()
        self.register_tray()

    def create_window(self) -> None:
        style = win32con.WS_OVERLAPPED | win32con.WS_SYSMENU
        window_name = "%s StatusIcon Window" % bytestostr(self.title)
        niwc = get_notifyicon_wnd_class()
        args = (0, niwc.NIclassAtom, window_name, style,
                win32con.CW_USEDEFAULT, win32con.CW_USEDEFAULT, 0, 0,
                0, 0, niwc.hInstance, 0)
        log("CreateWindowExA%s", args)
        self.hwnd = CreateWindowExA(*args)
        log("create_window() hwnd=%#x", self.hwnd or 0)
        if not self.hwnd:
            raise WinError(get_last_error())
        TASKBAR_CREATED = RegisterWindowMessageA(b"TaskbarCreated")
        log(f"{TASKBAR_CREATED=}")
        message_map[TASKBAR_CREATED] = win32NotifyIcon.OnTrayRestart
        UpdateWindow(self.hwnd)
        # register callbacks:
        win32NotifyIcon.instances[self.hwnd] = self

    def register_tray(self) -> None:
        ni = self.make_nid(NIF_ICON | NIF_MESSAGE | NIF_TIP)
        r = Shell_NotifyIcon(NIM_ADD, byref(ni))
        log("Shell_NotifyIcon ADD=%i", r)
        if not r:
            raise RuntimeError("Shell_NotifyIcon failed to ADD, is explorer.exe running?")

    def make_nid(self, flags) -> NOTIFYICONDATA:
        assert self.hwnd
        nid = NOTIFYICONDATA()
        nid.cbSize = sizeof(NOTIFYICONDATA)
        nid.hWnd = self.hwnd
        nid.uCallbackMessage = win32con.WM_MENUCOMMAND
        nid.hIcon = self.current_icon
        # don't ask why we have to use sprintf to get what we want:
        title = bytestostr(self.title[:MAX_TIP_SIZE - 1])
        try:
            nid.szTip = title
        except (TypeError, ValueError):
            nid.szTip = title.encode()
        nid.dwState = 0
        nid.dwStateMask = 0
        nid.guidItem = XPRA_GUID
        nid.uID = self.app_id
        flags |= NIF_GUID
        # balloon notification bits:
        # szInfo
        # uTimeout
        # szInfoTitle
        # dwInfoFlags
        # hBalloonIcon
        # flags |= NIF_SHOWTIP
        nid.uVersion = 4
        nid.uFlags = flags
        log("make_nid(..)=%s tooltip=%r, app_id=%i, actual flags=%s",
            nid, title, self.app_id, csv([v for k, v in NIF_FLAGS.items() if k & flags]))
        return nid

    def delete_tray_window(self) -> None:
        if not self.hwnd:
            return
        try:
            nid = NOTIFYICONDATA()
            nid.cbSize = sizeof(NOTIFYICONDATA)
            nid.hWnd = self.hwnd
            nid.uID = self.app_id
            nid.guidItem = XPRA_GUID
            nid.uFlags = NIF_GUID
            log("delete_tray_window(..) calling Shell_NotifyIcon(NIM_DELETE, %s)", nid)
            r = Shell_NotifyIcon(NIM_DELETE, byref(nid))
            log("Shell_NotifyIcon(NIM_DELETE, nid)=%s", bool(r))
            ci = self.current_icon
            di = self.destroy_icon
            if ci and di:
                self.current_icon = None
                self.destroy_icon = None
                di(ci)
        except Exception as e:
            log("delete_tray_window()", exc_info=True)
            log.error("Error: failed to delete tray window")
            log.estr(e)

    def get_geometry(self):
        # we can only use this if there is a single monitor
        # because multi-monitor coordinates may have offsets
        # we don't know about (done inside GTK)
        n = len(EnumDisplayMonitors())
        if n == 1:
            nii = NOTIFYICONIDENTIFIER()
            nii.cbSize = sizeof(NOTIFYICONIDENTIFIER)
            nii.hWnd = self.hwnd
            nii.uID = self.app_id
            # nii.guidItem = XPRA_GUID
            rect = RECT()
            if Shell_NotifyIconGetRect(byref(nii), byref(rect)) == 0:  # NOSONAR
                geom = (rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top)
                geomlog("Shell_NotifyIconGetRect: %s", geom)
                return geom
        return None

    def set_blinking(self, on) -> None:
        # implement blinking on win32 using a timer?
        pass

    def set_tooltip(self, tooltip: str) -> None:
        log("set_tooltip(%r)", tooltip)
        self.title = tooltip
        nid = self.make_nid(NIF_ICON | NIF_MESSAGE | NIF_TIP)
        Shell_NotifyIcon(NIM_MODIFY, byref(nid))

    def set_icon(self, iconPathName: str) -> None:
        log("set_icon(%s)", iconPathName)
        hicon = self.LoadImage(iconPathName) or FALLBACK_ICON
        self.do_set_icon(hicon)
        nid = self.make_nid(NIF_ICON)
        Shell_NotifyIcon(NIM_MODIFY, byref(nid))
        self.reset_function = self.set_icon, (iconPathName,)

    def do_set_icon(self, hicon, destroy_icon=None) -> None:
        log("do_set_icon(%#x)", hicon)
        ci = self.current_icon
        di = self.destroy_icon
        if ci and ci != hicon and di:
            di(ci)
        self.current_icon = hicon
        self.destroy_icon = destroy_icon
        nid = self.make_nid(NIF_ICON)
        Shell_NotifyIcon(NIM_MODIFY, byref(nid))

    def set_icon_from_data(self, pixels, has_alpha: bool, w: int, h: int, rowstride: int, options=None) -> None:
        # this is convoluted but it works..
        log("set_icon_from_data%s", ("%s pixels" % len(pixels), has_alpha, w, h, rowstride, options))
        from PIL import Image
        if has_alpha:
            img_format = "RGBA"
        else:
            img_format = "RGBX"
        rgb_format = typedict(options or {}).strget("rgb_format", "RGBA")
        img = Image.frombuffer(img_format, (w, h), pixels, "raw", rgb_format, rowstride, 1)
        assert img, "failed to load image from buffer (%i bytes for %ix%i %s)" % (len(pixels), w, h, rgb_format)
        # apparently, we have to use SM_CXSMICON (small icon) and not SM_CXICON (regular size):
        icon_w = GetSystemMetrics(win32con.SM_CXSMICON)
        icon_h = GetSystemMetrics(win32con.SM_CYSMICON)
        if w != icon_w or h != icon_h:
            log("resizing tray icon to %ix%i", icon_w, icon_h)
            try:
                LANCZOS = Image.Resampling.LANCZOS
            except AttributeError:
                LANCZOS = Image.LANCZOS
            img = img.resize((icon_w, icon_h), LANCZOS)
            rowstride = w * 4
        hicon = image_to_ICONINFO(img, TRAY_ALPHA) or FALLBACK_ICON
        self.do_set_icon(hicon, DestroyIcon)
        UpdateWindow(self.hwnd)
        self.reset_function = self.set_icon_from_data, (pixels, has_alpha, w, h, rowstride, options)

    def LoadImage(self, iconPathName: str):
        if not iconPathName:
            return None
        image = self.image_cache.get(iconPathName)
        if not image:
            image = self.doLoadImage(iconPathName)
            self.image_cache[iconPathName] = image
        return image

    def doLoadImage(self, iconPathName: str):
        mingw_prefix = os.environ.get("MINGW_PREFIX")
        if mingw_prefix and iconPathName.find(mingw_prefix) >= 0:
            # python can deal with mixed win32 and unix paths,
            # but the native win32 LoadImage function cannot,
            # this happens when running from a mingw environment
            iconPathName = iconPathName.replace("/", "\\")
        icon_flags = win32con.LR_LOADFROMFILE | win32con.LR_DEFAULTSIZE
        try:
            assert os.path.exists(iconPathName), "icon '%s' not found" % iconPathName
            img_type = win32con.IMAGE_ICON
            if iconPathName.lower().split(".")[-1] in ("png", "bmp"):
                img_type = win32con.IMAGE_BITMAP
                icon_flags |= win32con.LR_CREATEDIBSECTION | win32con.LR_LOADTRANSPARENT
            log("LoadImage(%s) using image type=%s", iconPathName,
                {
                    win32con.IMAGE_ICON: "ICON",
                    win32con.IMAGE_BITMAP: "BITMAP",
                }.get(img_type))
            niwc = get_notifyicon_wnd_class()
            v = LoadImageW(niwc.hInstance, iconPathName, img_type, 0, 0, icon_flags)
            assert v is not None
        except Exception:
            log.error("Error: failed to load icon '%s'", iconPathName, exc_info=True)
            return None
        else:
            log("LoadImage(%s)=%#x", iconPathName, v)
            return v

    def OnTrayRestart(self, hwnd=0, msg=0, wparam=0, lparam=0) -> None:
        rfn = self.reset_function
        try:
            # re-create the tray window:
            self.delete_tray_window()
            self.register_tray()
            # now try to repaint the tray:
            log("OnTrayRestart%s reset function: %s", (hwnd, msg, wparam, lparam), rfn)
            if rfn:
                rfn[0](*rfn[1])
        except Exception as e:
            log.error("Error: cannot reset tray icon")
            log.error(f" using {rfn}")
            log.estr(e)

    def OnCommand(self, hwnd: int, msg: int, wparam: int, lparam: int):
        cb = self.command_callback
        log("OnCommand%s callback=%s", (hwnd, msg, wparam, lparam), cb)
        if cb:
            cid = wparam & 0xFFFF
            cb(hwnd, cid)

    def OnDestroy(self, hwnd: int, msg: int, wparam: int, lparam: int) -> None:
        log("OnDestroy%s", (hwnd, msg, wparam, lparam))
        self.destroy()

    def OnTaskbarNotify(self, hwnd: int, msg: int, wparam: int, lparam: int) -> int:
        if lparam == win32con.WM_MOUSEMOVE:
            cb = self.move_callback
            bm = [(hwnd, int(msg), int(wparam), int(lparam))]
        else:
            cb = self.click_callback
            bm = BUTTON_MAP.get(lparam, [])
        log("OnTaskbarNotify%s button(s) lookup: %s, callback=%s", (hwnd, msg, wparam, lparam), bm, cb)
        for button_event in bm:
            cb(*button_event)
        return 1

    def close(self) -> None:
        log("win32NotifyIcon.close()")
        self.destroy()

    def destroy(self) -> None:
        cb = self.exit_callback
        hwnd = self.hwnd
        log("destroy() hwnd=%#x, exit callback=%s", hwnd, cb)
        self.delete_tray_window()
        self.exit_callback = noop
        with log.trap_error("Error on exit callback %s", cb):
            cb()
        if hwnd:
            win32NotifyIcon.instances.pop(hwnd, None)


WM_TRAY_EVENT = win32con.WM_MENUCOMMAND  # a message id we choose
log(f"{WM_TRAY_EVENT=}")

message_map: dict[int, Callable] = {
    win32con.WM_DESTROY: win32NotifyIcon.OnDestroy,
    win32con.WM_COMMAND: win32NotifyIcon.OnCommand,
    WM_TRAY_EVENT: win32NotifyIcon.OnTaskbarNotify,
}


def NotifyIconWndProc(hwnd, msg, wParam, lParam):
    instance = win32NotifyIcon.instances.get(hwnd)
    fn = message_map.get(msg)

    def i(v):
        try:
            return int(v)
        except (ValueError, TypeError):
            return v

    log("NotifyIconWndProc%s instance=%s, message(%i)=%s", (i(hwnd), i(msg), i(wParam), i(lParam)), instance, msg, fn)
    # log("potential matching win32 constants for message: %s", [x for x in dir(win32con) if getattr(win32con, x)==msg])
    if instance and fn:
        return fn(instance, hwnd, msg, wParam, lParam) or 0
    return DefWindowProcA(hwnd, msg, wParam, lParam)


_notifyicon_wnd_class = None


def get_notifyicon_wnd_class():
    global _notifyicon_wnd_class
    if _notifyicon_wnd_class is None:
        hmodule = HMODULE(0)
        assert GetModuleHandleExA(0, None, byref(hmodule))
        log("GetModuleHandleExA(..)=%#x", int(hmodule.value))

        NIwc = WNDCLASSEX()
        NIwc.cbSize = sizeof(WNDCLASSEX)
        NIwc.style = win32con.CS_HREDRAW | win32con.CS_VREDRAW
        NIwc.lpfnWndProc = WNDPROC(NotifyIconWndProc)
        NIwc.hInstance = hmodule
        NIwc.hBrush = GetStockObject(win32con.WHITE_BRUSH)
        NIwc.lpszClassName = "win32NotifyIcon"

        NIclassAtom = RegisterClassExA(byref(NIwc))
        log("RegisterClassExA(%s)=%i", NIwc.lpszClassName, NIclassAtom)
        if NIclassAtom == 0:
            raise WinError(get_last_error())
        NIwc.NIclassAtom = NIclassAtom
        _notifyicon_wnd_class = NIwc
    return _notifyicon_wnd_class


def main(args):
    from xpra.platform.win32.common import user32
    from ctypes.wintypes import MSG

    if "-v" in args:
        from xpra.log import enable_debug_for
        enable_debug_for("all")

    log.warn("main")

    def click_callback(_button, _pressed):
        menu = CreatePopupMenu()
        AppendMenu(menu, win32con.MF_STRING, 1024, "Generate balloon")
        AppendMenu(menu, win32con.MF_STRING, 1025, "Exit")
        pos = POINT()
        GetCursorPos(addressof(pos))  # NOSONAR
        hwnd = tray.hwnd
        user32.SetForegroundWindow(hwnd)
        user32.TrackPopupMenu(menu, win32con.TPM_LEFTALIGN, pos.x, pos.y, 0, hwnd, None)  # @UndefinedVariable
        PostMessageA(hwnd, win32con.WM_NULL, 0, 0)

    def command_callback(hwnd, cid):
        if cid == 1024:
            from xpra.platform.win32.balloon import notify
            try:
                from PIL import Image
                from io import BytesIO
                img = Image.open("icons\\printer.png")
                buf = BytesIO()
                img.save(buf, "PNG")
                data = buf.getvalue()
                buf.close()
                icon = (b"png", img.size[0], img.size[1], data)
            except Exception as e:
                print(f"could not find icon: {e}")
                icon = None
            notify(hwnd, 0, "hello", "world", timeout=1000, icon=icon)
        elif cid == 1025:
            print("Goodbye")
            DestroyWindow(hwnd)
        else:
            print("OnCommand for ID=%s" % cid)

    def win32_quit(*_args):
        PostQuitMessage(0)  # Terminate the app.

    from xpra.platform.paths import get_icon_dir, get_app_dir
    for idir in (get_icon_dir(), get_app_dir(), "fs/share/xpra/icons"):
        ipath = os.path.abspath(os.path.join(idir, "xpra.ico"))
        if os.path.exists(ipath):
            break
    tray = win32NotifyIcon(999, "test", move_callback=noop, click_callback=click_callback, exit_callback=win32_quit,
                           command_callback=command_callback, iconPathName=ipath)
    import signal
    signal.signal(signal.SIGINT, win32_quit)
    signal.signal(signal.SIGTERM, win32_quit)

    # pump messages:
    msg = MSG()
    p_msg = addressof(msg)  # NOSONAR
    while user32.GetMessageA(p_msg, win32con.NULL, 0, 0) != 0:
        user32.TranslateMessage(p_msg)
        user32.DispatchMessageA(p_msg)


if __name__ == '__main__':
    import sys

    main(sys.argv)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import socket
from zeroconf import ServiceBrowser, Zeroconf

from xpra.log import Logger

log = Logger("network", "mdns")


class Zeroconflistener:

    def __init__(self, service_type, mdns_found=None, mdns_add=None, mdns_remove=None, mdns_update=None):
        log("Zeroconflistener%s", (service_type, mdns_found, mdns_add, mdns_remove, mdns_update))
        self.zeroconf = Zeroconf()
        self.browser: ServiceBrowser | None = None
        if not service_type.endswith("local."):
            service_type += "local."
        self.service_type = service_type
        self.mdns_found = mdns_found
        self.mdns_add = mdns_add
        self.mdns_remove = mdns_remove
        self.mdns_update = mdns_update

    def __repr__(self):
        return "Zeroconflistener(%s)" % self.service_type

    def update_service(self, zeroconf, stype: str, name: str) -> None:
        log("update_service%s", (zeroconf, stype, name))
        if self.mdns_update:
            self.mdns_update(name, stype)

    def remove_service(self, zeroconf, stype: str, name: str) -> None:
        log("remove_service%s", (zeroconf, stype, name))
        if self.mdns_remove:
            domain = "local"
            self.mdns_remove(0, 0, name, stype, domain, 0)

    def add_service(self, zeroconf, stype: str, name: str):
        log("add_service%s", (zeroconf, stype, name))
        info = zeroconf.get_service_info(stype, name)
        log("service info: %s", info)
        if self.mdns_add and info:
            interface = None
            protocol = 0
            name = info.name
            stype = info.type
            domain = "local"
            server = info.server
            try:
                addresses = info.addresses
            except AttributeError:
                addresses = [info.address]
            port = info.port
            props = info.properties
            for address in addresses:
                saddress = socket.inet_ntoa(address)
                self.mdns_add(interface, protocol, name, stype, domain, server, saddress, port, props)

    def start(self) -> None:
        self.browser = ServiceBrowser(self.zeroconf, self.service_type, listener=self)
        log("ServiceBrowser%s=%s", (self.zeroconf, self.service_type, self), self.browser)

    def stop(self) -> None:
        b = self.browser
        if b:
            self.browser = None
            try:
                b.cancel()
            except Exception:
                pass
        zc = self.zeroconf
        if zc:
            self.zeroconf = None
            try:
                zc.close()
            except Exception:
                pass


def main():
    def mdns_found(*args):
        print(f"mdns_found: {args}")

    def mdns_add(*args):
        print(f"mdns_add: {args}")

    def mdns_remove(*args):
        print(f"mdns_remove: {args}")

    def mdns_update(*args):
        print(f"mdns_update: {args}")

    from xpra.os_util import gi_import
    GLib = gi_import("GLib")
    loop = GLib.MainLoop()

    from xpra.platform import program_context
    with program_context("zeroconf-listener", "zeroconf-listener"):
        listeners: list[Zeroconflistener] = []
        from xpra.net.mdns import XPRA_TCP_MDNS_TYPE, XPRA_UDP_MDNS_TYPE

        def add(service_type):
            listener = Zeroconflistener(service_type + "local.", mdns_found, mdns_add, mdns_remove, mdns_update)
            log(f"{listener=}")
            listener.start()

        add(XPRA_TCP_MDNS_TYPE)
        add(XPRA_UDP_MDNS_TYPE)
        try:
            loop.run()
        finally:
            for listener in listeners:
                listener.stop()


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Sequence

from xpra.auth.common import SessionData
from xpra.net.digest import get_salt, choose_digest
from xpra.util.objects import typedict


class Authenticator:
    def __init__(self, **kwargs):
        self.challenge_sent = False
        self.prompt: str = kwargs.pop("prompt", "password")
        self.passed: bool = False

    def requires_challenge(self) -> bool:
        return True

    def get_challenge(self, digests: Sequence[str]) -> tuple[bytes, str]:
        self.challenge_sent = True
        return get_salt(), choose_digest(digests)

    def choose_salt_digest(self, digest_modes: Sequence[str]) -> str:
        return choose_digest(digest_modes)

    def get_uid(self) -> int:
        return -1

    def get_gid(self) -> int:
        return -1

    def get_passwords(self) -> Sequence[str]:
        return ()

    def get_password(self) -> str:
        return ""

    def authenticate(self, _caps: typedict) -> bool:  # pylint: disable=unused-argument
        return False

    def get_sessions(self) -> SessionData | None:
        return None

    def __repr__(self):
        return "reject"



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import unittest

from xpra.util.str_fn import memoryview_to_bytes, hexstr, sorted_nicely
from xpra.util.objects import typedict
from xpra.codecs import loader
from xpra.codecs.image import ImageWrapper
from xpra.codecs.checks import make_test_image, TEST_COLORS


def cmp_bytes(b1: bytes, b2: bytes, tolerance=1) -> bool:
    length = min(len(b1), len(b2))
    for i in range(length):
        v1 = b1[i]
        v2 = b2[i]
        delta = abs(v2-v1)
        if delta > tolerance:
            print(f"pixel {i} /  {length}: {v1:x} vs {v2:x}, {delta:x}>{tolerance:x}")
            return False
    return True


def cmp_pixels(pix1: bytes, w1: int, h1: int, stride1: int, bpp1: int,
               pix2: bytes, w2: int, h2: int, stride2: int, bpp2: int,
               tolerance=1) -> bool:
    h = min(h1, h2)
    w = min(w1, w2)
    for y in range(h):
        row1 = pix1[y * stride1: y * stride1 + w * bpp1]
        row2 = pix2[y * stride2: y * stride2 + w * bpp2]
        if not cmp_bytes(row1, row2, tolerance):
            print("expected:")
            print(hexstr(row1))
            print("but got:")
            print(hexstr(row2))
            print(f"for row {y} / {h}")
            return False
    return True


def cmp_images(image1: ImageWrapper, image2: ImageWrapper, tolerance=1) -> bool:
    pixels1 = memoryview_to_bytes(memoryview(image1.get_pixels()))
    pixels2 = memoryview_to_bytes(memoryview(image2.get_pixels()))
    return cmp_pixels(
        pixels1, image1.get_width(), image1.get_height(), image1.get_rowstride(), image1.get_bytesperpixel(),
        pixels2, image2.get_width(), image2.get_height(), image2.get_rowstride(), image2.get_bytesperpixel(),
        tolerance)


def get_tolerance(enc_name: str, encoding: str, quality: int) -> int:
    tolerance = {
        100: 1 if encoding in ("jpeg",) else 0,
        90: 2 if encoding in ("webp",) else 1,
        50: 4,
        10: 0xc if encoding in ("webp",) else 4,
    }.get(quality, 0)
    if enc_name == "enc_pillow" and encoding == "webp":
        tolerance += 3
    return tolerance


class TestColorRange(unittest.TestCase):

    def setUp(self) -> None:
        self.width = 48
        self.height = 32
        self.rgb_format = "BGRX"
        self.test_info: set[str] = set()

    def tearDown(self) -> None:
        print(f"successfully tested {self.rgb_format} input with:")
        for s in sorted_nicely(self.test_info):
            print(f"{s}")

    def test_pseudo_video(self) -> None:
        self.do_test_pseudo_video("enc_webp", "webp")
        self.do_test_pseudo_video("enc_jpeg", "jpeg")

    def do_test_pseudo_video(self, enc_name: str, encoding: str) -> None:
        enc_mod = loader.load_codec(enc_name)
        for color, pixel in TEST_COLORS.items():
            image = make_test_image(self.rgb_format, self.width, self.height, pixel)
            self.pixels = image.get_pixels()
            for quality in (100, 90, 50, 10):
                tolerance = get_tolerance(enc_name, encoding, quality)
                options = typedict({
                    "quality": quality,
                    "color": color,
                    "tolerance": tolerance,
                })
                encoder = enc_mod.Encoder()
                encoder.init_context(encoding, self.width, self.height, self.rgb_format, typedict())
                data, client_options = encoder.compress_image(image, options=options)
                bdata = memoryview_to_bytes(data)
                self.verify_PIL(enc_name, encoding, image, options, bdata)
                self.verify_rgb(enc_name, encoding, image, options, bdata)
                self.verify_yuv(enc_name, encoding, image, options, bdata, client_options)

    def test_encode_decode_range(self) -> None:
        encoders = os.environ.get("XPRA_TEST_ENCODERS", ",".join(loader.ENCODER_CODECS)).split(",")
        for enc_name in encoders:
            enc_mod = loader.load_codec(enc_name)
            if not enc_mod:
                continue
            formats = enc_mod.get_encodings()
            for encoding in formats:
                for color, pixel in TEST_COLORS.items():
                    image = make_test_image(self.rgb_format, self.width, self.height, pixel)
                    # the encoder may modify the pixels value,
                    # (the spng encoder does)
                    self.pixels = image.get_pixels()
                    for quality in (100, 90, 50, 10):
                        tolerance = get_tolerance(enc_name, encoding, quality)
                        options = typedict({
                            "quality": quality,
                            "color": color,
                            "tolerance": tolerance,
                        })
                        output = enc_mod.encode(encoding, image, options=options)
                        # tuple[str, Compressed, dict[str, Any], int, int, int, int]
                        if not output:
                            raise RuntimeError(f"failed to encode {image} using {enc_mod.encode}")
                        bdata = memoryview_to_bytes(output[1].data)
                        client_options = output[2]
                        self.verify_PIL(enc_name, encoding, image, options, bdata)
                        self.verify_rgb(enc_name, encoding, image, options, bdata)
                        self.verify_yuv(enc_name, encoding, image, options, bdata, client_options)

    def verify_PIL(self, enc_name: str, encoding: str, image: ImageWrapper, enc_options: dict, data: bytes) -> None:
        ext = encoding.replace("/", "")  # ie: "png/L" -> "pngL"
        if ext not in ("png", "webp", "jpeg"):
            return
        # verify first 16 bytes when compressed image with Pillow:
        from io import BytesIO
        from PIL import Image
        img = Image.open(BytesIO(data))
        img = img.convert("RGBA")
        rdata = img.tobytes("raw", self.rgb_format.replace("X", "A"))
        tolerance = enc_options["tolerance"]
        bpp = len(self.rgb_format)
        # one pixel at a time:
        for pix_x in range(self.width):
            # test each component individually: 'R', 'G' and 'B'
            for i, component in enumerate(self.rgb_format):
                if component in ("X", "A"):
                    continue
                index = pix_x * bpp + i
                v1 = self.pixels[index]
                v2 = rdata[index]
                if abs(v1 - v2) > tolerance:
                    raise RuntimeError(f"pixels reloaded do not match: "
                                       f"expected {hexstr(self.pixels[:16])} but got {hexstr(rdata[:16])} "
                                       f"from {enc_name} using {encoding} and pixel format {self.rgb_format} "
                                       f"with {enc_options=} "
                                       f"mismatch at pixel {pix_x} for component {i}: {component}: {v1:x} vs {v2:x} "
                                       f"tolerance={tolerance}")
        self.test_info.add(f"{enc_name:12}  {encoding:12}  pillow                                    {self.rgb_format}")

    def verify_rgb(self, enc_name: str, encoding: str, image: ImageWrapper, enc_options: dict, data: bytes) -> None:
        # try to decompress to rgb:
        decoders = os.environ.get("XPRA_TEST_DECODERS", ",".join(loader.DECODER_CODECS)).split(",")
        for dec_name in decoders:
            dec_mod = loader.load_codec(dec_name)
            if not dec_mod:
                continue
            if encoding not in dec_mod.get_encodings():
                continue
            decompress_to_rgb = getattr(dec_mod, "decompress_to_rgb", None)
            # print(f"testing {fmt} rgb decoding using {decompress_to_rgb}")
            if not decompress_to_rgb:
                continue
            dec_options = typedict({"rgb_format": self.rgb_format})
            rimage = decompress_to_rgb(data, dec_options)
            tolerance = enc_options["tolerance"]
            if dec_name == "dec_jpeg":
                tolerance += 1
            if not cmp_images(image, rimage, tolerance):
                raise RuntimeError(f"decoder {dec_name} from {enc_name} produced an image that differs with {enc_options=}")
            self.test_info.add(f"{enc_name:12}  {encoding:12}  {dec_name:12}                              {self.rgb_format}")

    def verify_yuv(self, enc_name: str, encoding: str, image: ImageWrapper, enc_options: dict, data: bytes, client_options: dict) -> None:
        decoders = os.environ.get("XPRA_TEST_DECODERS", ",".join(loader.DECODER_CODECS)).split(",")
        csc_modules = os.environ.get("XPRA_TEST_CSC", ",".join(loader.CSC_CODECS)).split(",")
        # try to decompress to yuv
        for dec_name in decoders:
            dec_mod = loader.load_codec(dec_name)
            if not dec_mod:
                continue
            if encoding not in dec_mod.get_encodings():
                continue
            decompress_to_yuv = getattr(dec_mod, "decompress_to_yuv", None)
            if not decompress_to_yuv:
                continue
            dec_options = typedict(client_options)
            yuv_image = decompress_to_yuv(data, dec_options)
            assert yuv_image
            yuv_format = yuv_image.get_pixel_format()
            # find a csc module to convert this back to rgb:
            for csc_name in csc_modules:
                csc_mod = loader.load_codec(csc_name)
                if not csc_mod:
                    continue

                def find_spec():
                    for spec in csc_mod.get_specs():
                        if yuv_format not in spec.input_colorspace:
                            continue
                        if self.rgb_format not in spec.output_colorspaces:
                            continue
                        return spec
                    return None

                spec = find_spec()
                csc_options = typedict({"full-range": yuv_image.get_full_range()})
                converter = spec.codec_class()
                converter.init_context(self.width, self.height, yuv_format,
                                       self.width, self.height, self.rgb_format, csc_options)
                rgb_image = converter.convert_image(yuv_image)
                assert rgb_image
                tolerance = enc_options["tolerance"]
                if not cmp_images(image, rgb_image, tolerance + 2):
                    raise RuntimeError(f"decoder {dec_name} from {enc_name} produced a YUV image that differs with {enc_options=}"
                                       f" (converted to {self.rgb_format} from {yuv_format} using {csc_name})")
                self.test_info.add(f"{enc_name:12}  {encoding:12}  {dec_name:12}  {yuv_format:12}  {csc_name:12}  {self.rgb_format:12}")


def main():
    unittest.main()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.util.objects import AdHocStruct
from unit.process_test_util import DisplayContext
from unit.client.mixins.clientmixintest_util import ClientMixinTest


class WindowManagerTest(ClientMixinTest):

	def test_windowmanager(self):
		with DisplayContext():
			from xpra.client.mixins.windows import WindowClient
			def _WindowClient():
				def get_mouse_position():
					return 0, 0
				wc = WindowClient()
				wc.get_mouse_position = get_mouse_position
				return wc
			opts = AdHocStruct()
			opts.system_tray = True
			opts.cursors = True
			opts.bell = True
			opts.input_devices = True
			opts.auto_refresh_delay = 0
			opts.min_size = "100x100"
			opts.max_size = "2000x2000"
			opts.pixel_depth = 24
			opts.windows = True
			opts.window_close = "forward"
			opts.modal_windows = True
			opts.border = "red"
			opts.mousewheel = "yes"
			opts.tray_icon = "yes"
			self._test_mixin_class(_WindowClient, opts)

def main():
	unittest.main()


if __name__ == '__main__':
	main()



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

"""
The goo needed to deal with X properties.

Everyone else should just use prop_set/prop_get with nice clean Python calling
conventions, and if you need more (un)marshalling smarts, add them here.

This module adds GTK / GDK specific conversions,
the conversions for plain python types is found in prop_conv.py
"""

import struct

from xpra.x11.prop_conv import prop_encode, prop_decode, PROP_TYPES, PROP_SIZES
from xpra.x11.bindings.window import X11WindowBindings, PropertyError
from xpra.gtk.error import xsync, XError, XSyncContext
from xpra.util.str_fn import repr_ellipsized
from xpra.log import Logger

log = Logger("x11", "window")


def _get_atom(d) -> str | None:
    unpacked = struct.unpack(b"@L", d)[0]
    if unpacked == 0:
        log.warn("Warning: invalid zero atom value")
        return None
    with xsync:
        name = X11WindowBindings().get_atom_name(unpacked)
    if not name:
        log.error("invalid atom: %s - %s", repr(d), repr(unpacked))
        return None
    return name


def _get_xatom(str_or_int):
    with xsync:
        return X11WindowBindings().get_xatom(str_or_int)


PYTHON_TYPES: dict[str, str] = {
    "UTF8_STRING": "utf8",
    "STRING": "latin1",
    "ATOM": "atom",
    "CARDINAL": "u32",
    "INTEGER": "integer",
    "VISUALID": "visual",
    "WINDOW": "window",
}


def get_python_type(scalar_type: str) -> str:
    # ie: get_python_type("STRING") = "latin1"
    return PYTHON_TYPES.get(scalar_type, scalar_type)


def _to_atom(a) -> bytes:
    return struct.pack(b"@L", _get_xatom(a))


# add the GTK / GDK types to the conversion function list:
PROP_TYPES["atom"] = (str, "ATOM", 32, _to_atom, _get_atom, b"")


def prop_set(xid: int, key: str, etype: list | tuple | str, value) -> None:
    dtype, dformat, data = prop_encode(etype, value)
    raw_prop_set(xid, key, dtype, dformat, data)


def raw_prop_set(xid: int, key: str, dtype: str, dformat: int, data) -> None:
    if not isinstance(xid, int):
        raise TypeError(f"xid must be an int, not a {type(xid)}")
    with xsync:
        X11WindowBindings().XChangeProperty(xid, key, dtype, dformat, data)


def prop_type_get(xid: int, key: str):
    try:
        return X11WindowBindings().GetWindowPropertyType(xid, key)
    except XError:
        log("prop_type_get%s", (xid, key), exc_info=True)
        return None


# May return None.
def prop_get(xid: int, key: str, etype, ignore_errors: bool = False, raise_xerrors: bool = False):
    # ie: 0x4000, "_NET_WM_PID", "u32"
    if isinstance(etype, (list, tuple)):
        scalar_type = etype[0]
    else:
        scalar_type = etype  # ie: "u32"
    type_atom = PROP_TYPES[scalar_type][1]  # ie: "CARDINAL"
    buffer_size = PROP_SIZES.get(scalar_type, 65536)
    data = raw_prop_get(xid, key, type_atom, buffer_size, ignore_errors, raise_xerrors)
    if data is None:
        return None
    return do_prop_decode(key, etype, data, ignore_errors)


def raw_prop_get(xid: int, key: str, type_atom: str, buffer_size: int = 65536,
                 ignore_errors: bool = False, raise_xerrors: bool = False):
    if not isinstance(xid, int):
        raise TypeError(f"xid must be an int, not a {type(xid)}")
    try:
        with XSyncContext():
            data = X11WindowBindings().XGetWindowProperty(xid, key, type_atom, buffer_size)
        if data is None:
            if not ignore_errors:
                log("Missing property %s (%s)", key, type_atom)
            return None
    except XError:
        log("raw_prop_get%s", (xid, key, type_atom, ignore_errors, raise_xerrors), exc_info=True)
        if raise_xerrors:
            raise
        log.info(f"Missing window {xid:x} or wrong property type {key} ({type_atom})")
        return None
    except PropertyError as e:
        log("raw_prop_get%s", (xid, key, type_atom, ignore_errors, raise_xerrors), exc_info=True)
        if not ignore_errors:
            log.info(f"Missing property or wrong property type {key} ({type_atom})")
            log.info(f" on window {xid:x}")
            log.info(" %s", str(e) or type(e))
        return None
    return data


def _etypestr(etype) -> str:
    if isinstance(etype, (list, tuple)):
        scalar_type = etype[0]
        return f"array of {scalar_type}"
    return str(etype)


def do_prop_decode(key: str, etype, data, ignore_errors=False):
    try:
        with XSyncContext():
            return prop_decode(etype, data)
    except Exception:
        if ignore_errors:
            log("prop_get%s", (key, etype, ignore_errors), exc_info=True)
            return None
        log.warn("Error parsing property '%s' (%s)", key, _etypestr(etype))
        log.warn(" this may be a misbehaving application, or bug in Xpra")
        try:
            log.warn(" data length=%i", len(data))
        except TypeError:
            pass
        log.warn(" data: %r", repr_ellipsized(str(data)), exc_info=True)
        raise


def prop_del(xid: int, key: str):
    if not isinstance(xid, int):
        raise TypeError(f"xid must be an int, not a {type(xid)}")
    with xsync:
        X11WindowBindings().XDeleteProperty(xid, key)



# This file is part of Xpra.
# Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os.path
import sys
import tempfile
import platform
from ctypes import WinDLL, create_unicode_buffer
from ctypes.wintypes import MAX_PATH

from xpra.util.io import get_util_logger

shell32 = WinDLL("shell32", use_last_error=True)
SHGetFolderPath = shell32.SHGetFolderPathW

CSIDL_APPDATA = 26
CSIDL_LOCAL_APPDATA = 28
CSIDL_COMMON_APPDATA = 35
CSIDL_PROFILE = 40


def sh_get_folder_path(v) -> str:
    try:
        buf = create_unicode_buffer(MAX_PATH)
        SHGetFolderPath(0, v, None, 0, buf)
        if not buf.value:
            return buf.value
        return os.path.normpath(buf.value)
    except Exception:
        return ""


def _get_data_dir(roaming=True) -> str:
    # if not running from a binary, return current directory:
    if not getattr(sys, 'frozen', ''):
        return os.getcwd()
    return get_appdata_dir(roaming)


def get_appdata_dir(roaming=True) -> str:
    appdata = sh_get_folder_path(CSIDL_APPDATA if roaming else CSIDL_LOCAL_APPDATA)
    if not appdata:
        # on win32 we must send stdout to a logfile to prevent an alert box on exit shown by `py2exe`
        # UAC in vista onwards will not allow us to write where the software is installed,
        # so we place the log file (etc) in "~/Application Data"
        appdata = os.environ.get("APPDATA" if roaming else "LOCALAPPDATA", "")
    if not appdata:
        # we need some kind of path..
        appdata = tempfile.gettempdir()
        assert appdata, "cannot find any usable directory for log files"
    if not os.path.exists(appdata):
        os.mkdir(appdata)
    data_dir = os.path.join(appdata, "Xpra")
    return data_dir


def do_get_resources_dir() -> str:
    from xpra.platform.paths import get_app_dir
    app_dir = get_app_dir()
    prefix = os.environ.get("MINGW_PREFIX")
    for d in (app_dir, prefix):
        if not d or not os.path.isdir(d):
            continue
        share_xpra = os.path.join(d, "share", "xpra")
        if os.path.exists(share_xpra):
            return share_xpra
        # source directory?
        share_xpra = os.path.join(d, "fs", "share", "xpra")
        if os.path.exists(share_xpra):
            return share_xpra
    return app_dir


def do_get_image_dir() -> str:
    from xpra.platform.paths import get_resources_dir
    return os.path.join(get_resources_dir(), "images")


def do_get_icon_dir() -> str:
    from xpra.platform.paths import get_resources_dir
    return os.path.join(get_resources_dir(), "icons")


def do_get_default_log_dirs() -> list[str]:
    dd = _get_data_dir()
    temp = tempfile.gettempdir()
    if dd == temp:
        return [temp]
    return [dd, temp]


def get_program_data_dir() -> str:
    # ie: "C:\ProgramData"
    try:
        return sh_get_folder_path(CSIDL_COMMON_APPDATA) or "C:\\ProgramData"
    except Exception:
        get_util_logger().debug("get_program_data_dir()", exc_info=True)
    return "C:\\ProgramData"


def do_get_system_conf_dirs() -> list[str]:
    # ie: C:\ProgramData\Xpra
    return [os.path.join(get_program_data_dir(), "Xpra")]


def do_get_system_menu_dirs() -> list[str]:
    from xpra.platform.win32 import get_common_startmenu_dir, get_startmenu_dir
    menu_dirs = []
    for menu_dir in (get_common_startmenu_dir(), get_startmenu_dir()):
        if not menu_dir or not os.path.exists(menu_dir) or not os.path.isdir(menu_dir):
            continue
        # ie: "C:\ProgramData\Microsoft\Windows\Start Menu"
        menu_dirs.append(menu_dir)
    return menu_dirs


def do_get_ssl_cert_dirs() -> list[str]:
    dirs = []
    for i in (CSIDL_PROFILE, CSIDL_COMMON_APPDATA, CSIDL_LOCAL_APPDATA, CSIDL_APPDATA):
        fpath = sh_get_folder_path(i)
        if not fpath:
            continue
        d = os.path.join(fpath, "Xpra")
        dirs.append(d)
        d = os.path.join(fpath, "Xpra", "ssl")
        dirs.append(d)
    dirs += do_get_default_conf_dirs()
    return dirs


def do_get_ssh_conf_dirs() -> list[str]:
    if platform.architecture()[0] == "32bit":
        system32 = "SysNative"
    else:
        system32 = "System32"
    windows_dir = os.environ.get("SystemRoot", os.environ.get("WINDIR", "C:\\Windows"))
    openssh_dir = os.path.join(windows_dir, system32, "OpenSSH")
    dirs = []
    for i in (CSIDL_PROFILE, CSIDL_COMMON_APPDATA, CSIDL_LOCAL_APPDATA, CSIDL_APPDATA):
        fpath = sh_get_folder_path(i)
        if not fpath:
            continue
        d = os.path.join(fpath, "SSH")
        dirs.append(d)
    dirs += do_get_default_conf_dirs() + [
        openssh_dir,  # ie: C:\Windows\system32\OpenSSH
        "~/.ssh",
        "~/ssh",
    ]
    return dirs


def do_get_ssh_known_hosts_files() -> list[str]:
    # reverse the order (avoid dotfiles on win32):
    return ["~/ssh/known_hosts", "~/.ssh/known_hosts"]


def do_get_state_dir() -> str:
    return "%APPDATA%\\Xpra\\state"


def do_get_sessions_dir() -> str:
    return "%APPDATA%\\Xpra\\sessions"


def do_get_default_conf_dirs() -> list[str]:
    # ie: C:\Program Files\Xpra\
    from xpra.platform.paths import get_app_dir
    return [os.path.join(get_app_dir(), "etc", "xpra")]


def do_get_user_conf_dirs(_uid) -> list[str]:
    dd = _get_data_dir()
    # ie: "C:\Users\<user name>\AppData\Roaming"
    SYSTEMROOT = os.environ.get("SYSTEMROOT", "")
    # ie: when running as a system service, we may get:
    # "C:\Windows\System32\config\systemprofile\AppData\Roaming"
    # and we don't want to use that:
    if SYSTEMROOT and dd.startswith(SYSTEMROOT):
        return []
    if not os.path.exists(dd):
        os.mkdir(dd)
    return [dd]


def do_get_desktop_background_paths() -> list[str]:
    try:
        from winreg import OpenKey, HKEY_CURRENT_USER, KEY_READ, QueryValueEx  # @Reimport
        key_path = "Control Panel\\Desktop"
        key = OpenKey(HKEY_CURRENT_USER, key_path, 0, KEY_READ)
        wallpaper = QueryValueEx(key, 'WallPaper')[0]
        return [wallpaper, ]
    except Exception:
        log = get_util_logger()
        log("do_get_desktop_background_paths()", exc_info=True)
    return []


def do_get_download_dir() -> str:
    try:
        # values found here: https://stackoverflow.com/a/48706260
        from winreg import OpenKey, HKEY_CURRENT_USER, QueryValueEx  # @Reimport
        sub_key = r'SOFTWARE\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders'
        downloads_guid = '{374DE290-123F-4565-9164-39C4925E467B}'
        with OpenKey(HKEY_CURRENT_USER, sub_key) as key:
            DOWNLOAD_PATH = QueryValueEx(key, downloads_guid)[0]
    except Exception:
        get_util_logger()("do_get_download_dir()", exc_info=True)
        # fallback to what the documentation says is the default:
        DOWNLOAD_PATH = os.path.join(os.environ.get("USERPROFILE", "~"), "My Documents", "Downloads")
        if not os.path.exists(DOWNLOAD_PATH):
            DOWNLOAD_PATH = os.path.join(os.environ.get("USERPROFILE", "~"), "Downloads")
    return DOWNLOAD_PATH


def do_get_script_bin_dirs() -> list[str]:
    # we don't save the "run-xpra" script anywhere on win32
    return []


def do_get_socket_dirs() -> list[str]:
    return []


APP_DIR: str = ""

if getattr(sys, 'frozen', False) is True:
    # cx_freeze = sys.frozen == True
    APP_DIR = os.path.dirname(sys.executable)
    if len(APP_DIR) > 3 and APP_DIR[1] == ":" and APP_DIR[2] == "/":
        # it seems that mingw builds can get confused about the correct value for os.pathsep:
        APP_DIR = APP_DIR.replace("/", "\\")
    try:
        sys.path.remove(APP_DIR)
    except ValueError:
        pass
    sys.path.insert(0, APP_DIR)
    os.chdir(APP_DIR)
    # so we can easily load DLLs with ctypes:
    os.environ['PATH'] = APP_DIR + os.pathsep + os.environ['PATH']


def do_get_app_dir() -> str:
    global APP_DIR
    if APP_DIR:
        return APP_DIR
    from xpra.platform.paths import default_get_app_dir  # imported here to prevent import loop
    return default_get_app_dir()


def do_get_nodock_command() -> list[str]:
    return _get_xpra_exe_command(
        "Xpra",  # executable without a shell
        "Xpra_cmd",  # we should never end up using this one
    )


def do_get_audio_command() -> list[str]:
    return _get_xpra_exe_command(
        "Xpra_Audio",  # executable without a shell, and with a nicer name
        "Xpra",  # executable without a shell
        "Xpra_cmd",  # we should never end up using this one
    )


def _get_xpra_exe_command(*cmd_options: str) -> list[str]:
    from xpra.platform.paths import get_app_dir
    exe_dir = get_app_dir()
    mingw = os.environ.get("MINGW_PREFIX")
    for cmd in cmd_options:
        exe_name = "%s.exe" % cmd
        if sys.executable.lower().endswith(exe_name):
            # prefer the same executable we were launched from:
            return [sys.executable]
        # try to find it in exe dir:
        exe = os.path.join(exe_dir, exe_name)
        if os.path.exists(exe) and os.path.isfile(exe):
            return [exe]
        # without ".exe" extension:
        script: str = os.path.join(exe_dir, cmd)
        if os.path.exists(script) and os.path.isfile(script):
            return [script]
        if mingw:
            # the python interpreter to use with the scripts:
            py = os.path.join(mingw, "bin", "python3.exe")
            if os.path.exists(py):
                if cmd.lower() in ("python", "python3"):
                    return [py]
                # ie: /e/Xpra/trunk/src/dist/xpra
                script = os.path.join(os.getcwd(), "scripts", cmd.lower())
                if os.path.exists(script) and os.path.isfile(script):
                    return [py, script]
                # ie: /mingw64/bin/xpra
                script = os.path.join(mingw, "bin", cmd.lower())
                if os.path.exists(script) and os.path.isfile(script):
                    return [py, script]
    from xpra.platform.paths import default_do_get_xpra_command
    d = default_do_get_xpra_command()
    if len(d) == 1:
        # ie: d="xpra"
        if not d[0].lower().endswith(".exe"):
            return [sys.executable, d[0]]
    return d


def do_get_xpra_command() -> list[str]:
    sl = sys.executable.lower()
    # keep the exact same command used to launch if we can:
    if sl.endswith("xpra_cmd.exe") or sl.endswith("xpra.exe"):
        return [sys.executable]
    return _get_xpra_exe_command("Xpra", "Xpra_cmd")


def do_get_python_exec_command() -> list[str]:
    return _get_xpra_exe_command("Python_exec_gui", "Python")


def do_get_python_execfile_command() -> list[str]:
    return _get_xpra_exe_command("Python_execfile_gui", "Python")



# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


from collections.abc import Callable

from xpra.challenge.handler import AuthenticationHandler


PROMPT_FN_KEY = "challenge_prompt_function"


def noprompt(prompt: str) -> str:
    return ""


class Handler(AuthenticationHandler):

    def __init__(self, **kwargs):
        self.challenge_prompt_fn: Callable[[str], str] = kwargs.get(PROMPT_FN_KEY, noprompt)

    def __repr__(self):
        return "prompt"

    def get_digest(self) -> str:
        return ""

    def handle(self, challenge: bytes, digest: str, prompt: str = "password"):  # pylint: disable=unused-argument
        digest_type = digest.split(":", 1)[0]
        if not prompt and digest_type in ("gss", "kerberos"):
            prompt = f"{digest_type} token"
        return self.challenge_prompt_fn(prompt)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>

from xpra.x11.gtk.display_source import init_gdk_display_source

init_gdk_display_source()
from xpra.x11.bindings.keyboard import X11KeyboardBindings

keyboard_bindings = X11KeyboardBindings()


def main():
    mappings = keyboard_bindings.get_keycode_mappings()
    print("mappings=%s" % mappings)
    print("")
    for k, v in mappings.items():
        print("%s\t\t:\t%s" % (k, v))


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys

from xpra.util.env import envbool
from xpra.os_util import WIN32, OSX

XPRA_TCP_MDNS_TYPE = "_xpra._tcp."
XPRA_UDP_MDNS_TYPE = "_xpra._udp."
RFB_MDNS_TYPE = "_rfb._tcp."

AVAHI = envbool("XPRA_AVAHI", not (WIN32 or OSX))
ZEROCONF = envbool("XPRA_ZEROCONF", True)
# publishes the name of the interface we broadcast from:
SHOW_INTERFACE = envbool("XPRA_MDNS_SHOW_INTERFACE", True)


def get_listener_class() -> type | None:
    from xpra.log import Logger
    log = Logger("mdns")
    log("mdns.get_listener_class() AVAHI=%s, ZEROCONF=%s", AVAHI, ZEROCONF)
    if ZEROCONF:
        # workaround for macOS Big Sur which broke ctypes,
        # ctypes is used in the ifaddr module which is imported by zeroconf:
        if sys.platform.startswith("darwin"):
            import xpra.platform  # pylint: disable=import-outside-toplevel
            # on macOS, an import side effect is to patch the ctypes loader
            assert xpra.platform
        try:
            from xpra.net.mdns.zeroconf_listener import Zeroconflistener
            log("Zeroconflistener=%s", Zeroconflistener)
            return Zeroconflistener
        except (ImportError, OSError):
            log("failed to import Zeroconflistener", exc_info=True)
    if AVAHI:
        try:
            from xpra.net.mdns.avahi_listener import Avahilistener
            log("Avahilistener=%s", Avahilistener)
            return Avahilistener
        except ImportError:
            log("failed to import Avahilistener", exc_info=True)
    log("mdns.get_listener_class()=None no backend found")
    return None



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

cdef memoryview argbdata_to_rgba(const unsigned int* argb, const int argb_len)  #pylint: disable=syntax-error
cdef memoryview argbdata_to_rgb(const unsigned int* argb, const int argb_len)
cdef memoryview bgradata_to_rgb(const unsigned int* bgra, const int bgra_len)
cdef memoryview bgradata_to_rgba(const unsigned int* bgra, const int bgra_len)

cdef void show_plane_range(name, plane: Buffer, int width, int stride, int height)



#!/usr/bin/env python3

import os

import gi
gi.require_version("Gtk", "3.0")  # @UndefinedVariable
from gi.repository import Gtk  # @UnresolvedImport


def main():
	window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
	window.set_size_request(400, 200)
	window.connect("delete_event", Gtk.main_quit)
	for x in ("/usr/share/icons/gnome/48x48/emblems/emblem-important.png", "/opt/share/icons/xpra.png"):
		if os.path.exists(x):
			print("using %s" % x)
			window.set_icon_from_file(x)
			break
	window.show_all()
	Gtk.main()
	return 0


if __name__ == "__main__":
	main()



#!/usr/bin/env python3
# This file is part of Xpra.
# This code is released under the terms of the MIT license:
# https://opensource.org/licenses/MIT
# This LSA code was found here:
# https://stackoverflow.com/questions/43114209/

import os
import sys
from ctypes import (
    Structure, POINTER,
    get_last_error, WinError, WinDLL,
    create_unicode_buffer, create_string_buffer, resize, addressof, byref, sizeof, memmove, cast,
    c_char, c_wchar, c_ulonglong, c_size_t, c_void_p,
)
from ctypes.wintypes import ULONG, LONG, BOOL, LARGE_INTEGER, USHORT, DWORD, LPVOID, LPWSTR, HANDLE
from collections import namedtuple

from xpra.log import Logger

log = Logger("win32")

ntdll = WinDLL('ntdll')
secur32 = WinDLL('secur32')
kernel32 = WinDLL('kernel32', use_last_error=True)
advapi32 = WinDLL('advapi32', use_last_error=True)

CHAR = c_char
WCHAR = c_wchar
PCHAR = POINTER(CHAR)
PWCHAR = POINTER(WCHAR)
SIZE_T = c_size_t
PULONG = POINTER(ULONG)
LSA_OPERATIONAL_MODE = ULONG
PLSA_OPERATIONAL_MODE = PULONG
PHANDLE = POINTER(HANDLE)
PLPVOID = POINTER(LPVOID)
LPDWORD = POINTER(DWORD)

CloseHandle = kernel32.CloseHandle
CloseHandle.argtypes = [HANDLE]

MAX_COMPUTER_NAME_LENGTH = 15

SECURITY_LOGON_TYPE = ULONG
Interactive = 2
Network = 3
Batch = 4
Service = 5

LOGON_SUBMIT_TYPE = ULONG
PROFILE_BUFFER_TYPE = ULONG

MsV1_0InteractiveLogon = 2
MsV1_0Lm20Logon = 3
MsV1_0NetworkLogon = 4
MsV1_0WorkstationUnlockLogon = 7
MsV1_0S4ULogon = 12
MsV1_0NoElevationLogon = 82

KerbInteractiveLogon = 2
KerbWorkstationUnlockLogon = 7
KerbS4ULogon = 12

MSV1_0_S4U_LOGON_FLAG_CHECK_LOGONHOURS = 0x2

KERB_S4U_LOGON_FLAG_CHECK_LOGONHOURS = 0x2
KERB_S4U_LOGON_FLAG_IDENTITY = 0x8

TOKEN_SOURCE_LENGTH = 8

NEGOTIATE_PACKAGE_NAME = b'Negotiate'
MICROSOFT_KERBEROS_NAME = b'Kerberos'
MSV1_0_PACKAGE_NAME = b'MICROSOFT_AUTHENTICATION_PACKAGE_V1_0'

DELETE = 0x00010000
READ_CONTROL = 0x00020000
WRITE_DAC = 0x00040000
WRITE_OWNER = 0x00080000

STANDARD_RIGHTS_REQUIRED = DELETE | READ_CONTROL | WRITE_DAC | WRITE_OWNER

TOKEN_ASSIGN_PRIMARY = 0x0001
TOKEN_DUPLICATE = 0x0002
TOKEN_IMPERSONATE = 0x0004
TOKEN_QUERY = 0x0008
TOKEN_QUERY_SOURCE = 0x0010
TOKEN_ADJUST_PRIVILEGES = 0x0020
TOKEN_ADJUST_GROUPS = 0x0040
TOKEN_ADJUST_DEFAULT = 0x0080
TOKEN_ADJUST_SESSIONID = 0x0100

TOKEN_ALL_ACCESS = (
    STANDARD_RIGHTS_REQUIRED | TOKEN_ASSIGN_PRIMARY | TOKEN_DUPLICATE | TOKEN_IMPERSONATE |
    TOKEN_QUERY | TOKEN_QUERY_SOURCE | TOKEN_ADJUST_PRIVILEGES | TOKEN_ADJUST_GROUPS |
    TOKEN_ADJUST_DEFAULT | TOKEN_ADJUST_SESSIONID
)

DUPLICATE_CLOSE_SOURCE = 0x00000001
DUPLICATE_SAME_ACCESS = 0x00000002

TOKEN_TYPE = ULONG
TokenPrimary = 1
TokenImpersonation = 2

SECURITY_IMPERSONATION_LEVEL = ULONG
SecurityAnonymous = 0
SecurityIdentification = 1
SecurityImpersonation = 2
SecurityDelegation = 3


class NTSTATUS(LONG):

    def to_error(self):
        return ntdll.RtlNtStatusToDosError(self)

    def __repr__(self):
        name = self.__class__.__name__
        status = ULONG.from_buffer(self)
        return '{}({:#010x})'.format(name, status.value)


PNTSTATUS = POINTER(NTSTATUS)


class MANAGED_HANDLE(HANDLE):
    __slots__ = 'closed',

    def __int__(self):
        return self.value or 0

    def detach(self):
        if not getattr(self, 'closed', False):
            self.closed = True
            value = int(self)
            self.value = None
            return value
        raise ValueError("already closed")

    def close(self):
        if self and not getattr(self, 'closed', False):
            CloseHandle(self.detach())

    __del__ = close

    def __repr__(self):
        return "%s(%d)" % (self.__class__.__name__, int(self))


class LARGE_INTEGER_TIME(LARGE_INTEGER):
    # https://msdn.microsoft.com/en-us/library/ff553204
    ntdll.RtlSecondsSince1970ToTime.restype = None
    _unix_epoch_LI = LARGE_INTEGER()
    ntdll.RtlSecondsSince1970ToTime(0, byref(_unix_epoch_LI))
    _unix_epoch = _unix_epoch_LI.value

    def __int__(self):
        return self.value

    def __repr__(self):
        name = self.__class__.__name__
        return '%s(%d)' % (name, self.value)

    def as_time(self) -> float:
        time100ns = self.value - self._unix_epoch
        if time100ns >= 0:
            return time100ns / 1e7
        raise ValueError('value predates the Unix epoch')

    @classmethod
    def from_time(cls, t):
        time100ns = int(t * 10 ** 7)
        return cls(time100ns + cls._unix_epoch)


class STRING(Structure):
    _fields_ = (
        ('Length', USHORT),
        ('MaximumLength', USHORT),
        ('Buffer', PCHAR),
    )


PSTRING = POINTER(STRING)


class UNICODE_STRING(Structure):
    _fields_ = (
        ('Length', USHORT),
        ('MaximumLength', USHORT),
        ('Buffer', PWCHAR),
    )


PUNICODE_STRING = POINTER(UNICODE_STRING)


class LUID(Structure):
    _fields_ = (
        ('LowPart', DWORD),
        ('HighPart', LONG),
    )

    def __new__(cls, value=0):
        return cls.from_buffer_copy(c_ulonglong(value))

    def __int__(self):
        return c_ulonglong.from_buffer(self).value

    def __repr__(self):
        name = self.__class__.__name__
        return '%s(%#x)' % (name, int(self))


PLUID = POINTER(LUID)
PSID = LPVOID


class SID_AND_ATTRIBUTES(Structure):
    _fields_ = (
        ('Sid', PSID),
        ('Attributes', DWORD),
    )


PSID_AND_ATTRIBUTES = POINTER(SID_AND_ATTRIBUTES)


class TOKEN_GROUPS(Structure):
    _fields_ = (
        ('GroupCount', DWORD),
        ('Groups', SID_AND_ATTRIBUTES * 1),
    )


PTOKEN_GROUPS = POINTER(TOKEN_GROUPS)


class TOKEN_SOURCE(Structure):
    # noinspection PyTypeChecker
    _fields_ = (
        ('SourceName', CHAR * TOKEN_SOURCE_LENGTH),
        ('SourceIdentifier', LUID),
    )

    def __init__(self, SourceName=None, SourceIdentifier=None):
        super().__init__()
        if SourceName is not None:
            if not isinstance(SourceName, bytes):
                SourceName = SourceName.encode('mbcs')
            self.SourceName = SourceName
        if SourceIdentifier is None:
            luid = self.SourceIdentifier
            ntdll.NtAllocateLocallyUniqueId(byref(luid))
        else:
            self.SourceIdentifier = SourceIdentifier


PTOKEN_SOURCE = POINTER(TOKEN_SOURCE)

py_source_context = TOKEN_SOURCE(b"PYTHON  ")
py_origin_name = b"Python-%d" % os.getpid()
py_logon_process_name = b"PythonLogonProcess-%d" % os.getpid()


class QUOTA_LIMITS(Structure):
    _fields_ = (
        ('PagedPoolLimit', SIZE_T),
        ('NonPagedPoolLimit', SIZE_T),
        ('MinimumWorkingSetSize', SIZE_T),
        ('MaximumWorkingSetSize', SIZE_T),
        ('PagefileLimit', SIZE_T),
        ('TimeLimit', LARGE_INTEGER_TIME),
    )


PQUOTA_LIMITS = POINTER(QUOTA_LIMITS)


class ContiguousUnicode(Structure):
    # _string_names_: sequence matched to underscore-prefixed fields

    def _get_unicode_string(self, name):
        wchar_size = sizeof(WCHAR)
        s = getattr(self, '_%s' % name)
        length = s.Length // wchar_size
        buf = s.Buffer
        if buf:
            return buf[:length]
        return None

    def _set_unicode_buffer(self, value):
        cls = type(self)
        wchar_size = sizeof(WCHAR)
        bufsize = (len(value) + 1) * wchar_size
        resize(self, sizeof(cls) + bufsize)
        addr = addressof(self) + sizeof(cls)
        src_buf = create_unicode_buffer(value)
        memmove(addr, addressof(src_buf), bufsize)  # NOSONAR

    def _set_unicode_string(self, name, value):
        values = []
        for n in self._string_names_:
            if n == name:
                values.append(value or '')
            else:
                values.append(getattr(self, n) or '')
        self._set_unicode_buffer('\x00'.join(values))

        cls = type(self)
        wchar_size = sizeof(WCHAR)
        addr = addressof(self) + sizeof(cls)
        for n, v in zip(self._string_names_, values):
            ptr = cast(addr, PWCHAR)
            ustr = getattr(self, '_%s' % n)
            length = ustr.Length = len(v) * wchar_size
            full_length = length + wchar_size
            if (n == name and value is None) or (n != name and not (length or ustr.Buffer)):
                ustr.Buffer = None
                ustr.MaximumLength = 0
            else:
                ustr.Buffer = ptr
                ustr.MaximumLength = full_length
            addr += full_length

    def __getattr__(self, name):
        if name not in self._string_names_:
            raise AttributeError
        return self._get_unicode_string(name)

    def __setattr__(self, name, value):
        if name in self._string_names_:
            self._set_unicode_string(name, value)
        else:
            super().__setattr__(name, value)

    @classmethod
    def from_address_copy(cls, address, size=None):
        x = Structure.__new__(cls)
        if size is not None:
            resize(x, size)
        memmove(byref(x), address, sizeof(x))
        delta = addressof(x) - address
        for n in cls._string_names_:
            ustr = getattr(x, '_%s' % n)
            addr = c_void_p.from_buffer(ustr.Buffer)
            if addr:
                addr.value += delta
        return x


class AuthInfo(ContiguousUnicode):
    # _message_type_: from a logon-submit-type enumeration
    def __init__(self):
        super().__init__()
        self.MessageType = self._message_type_


class MSV1_0_INTERACTIVE_LOGON(AuthInfo):
    _message_type_ = MsV1_0InteractiveLogon
    _string_names_ = 'LogonDomainName', 'UserName', 'Password'

    _fields_ = (
        ('MessageType', LOGON_SUBMIT_TYPE),
        ('_LogonDomainName', UNICODE_STRING),
        ('_UserName', UNICODE_STRING),
        ('_Password', UNICODE_STRING),
    )

    def __init__(self, UserName=None, Password=None, LogonDomainName=None):
        super().__init__()
        if LogonDomainName is not None:
            self.LogonDomainName = LogonDomainName
        if UserName is not None:
            self.UserName = UserName
        if Password is not None:
            self.Password = Password


class S4ULogon(AuthInfo):
    _string_names_ = 'UserPrincipalName', 'DomainName'

    _fields_ = (
        ('MessageType', LOGON_SUBMIT_TYPE),
        ('Flags', ULONG),
        ('_UserPrincipalName', UNICODE_STRING),
        ('_DomainName', UNICODE_STRING),
    )

    def __init__(self, UserPrincipalName=None, DomainName=None, Flags=0):
        super().__init__()
        self.Flags = Flags
        if UserPrincipalName is not None:
            self.UserPrincipalName = UserPrincipalName
        if DomainName is not None:
            self.DomainName = DomainName


class MSV1_0_S4U_LOGON(S4ULogon):
    _message_type_ = MsV1_0S4ULogon


class KERB_S4U_LOGON(S4ULogon):
    _message_type_ = KerbS4ULogon


PMSV1_0_S4U_LOGON = POINTER(MSV1_0_S4U_LOGON)
PKERB_S4U_LOGON = POINTER(KERB_S4U_LOGON)


class ProfileBuffer(ContiguousUnicode):
    # _message_type_
    def __init__(self):
        super().__init__()
        self.MessageType = self._message_type_


class MSV1_0_INTERACTIVE_PROFILE(ProfileBuffer):
    _message_type_ = MsV1_0InteractiveLogon
    _string_names_ = ('LogonScript', 'HomeDirectory', 'FullName',
                      'ProfilePath', 'HomeDirectoryDrive', 'LogonServer')
    _fields_ = (
        ('MessageType', PROFILE_BUFFER_TYPE),
        ('LogonCount', USHORT),
        ('BadPasswordCount', USHORT),
        ('LogonTime', LARGE_INTEGER_TIME),
        ('LogoffTime', LARGE_INTEGER_TIME),
        ('KickOffTime', LARGE_INTEGER_TIME),
        ('PasswordLastSet', LARGE_INTEGER_TIME),
        ('PasswordCanChange', LARGE_INTEGER_TIME),
        ('PasswordMustChange', LARGE_INTEGER_TIME),
        ('_LogonScript', UNICODE_STRING),
        ('_HomeDirectory', UNICODE_STRING),
        ('_FullName', UNICODE_STRING),
        ('_ProfilePath', UNICODE_STRING),
        ('_HomeDirectoryDrive', UNICODE_STRING),
        ('_LogonServer', UNICODE_STRING),
        ('UserFlags', ULONG),
    )

    def __repr__(self):
        return f"<MSV1_0_INTERACTIVE_PROFILE>({self.FullName!r})"


class SECURITY_ATTRIBUTES(Structure):
    _fields_ = (
        ('nLength', DWORD),
        ('lpSecurityDescriptor', LPVOID),
        ('bInheritHandle', BOOL),
    )

    def __init__(self, **kwds):
        self.nLength = sizeof(self)
        super().__init__(**kwds)


LPSECURITY_ATTRIBUTES = POINTER(SECURITY_ATTRIBUTES)


def _check_status(result, func, args):
    if result.value < 0:
        raise WinError(result.to_error())
    return args


def _check_bool(result, func, args):
    if not result:
        raise WinError(get_last_error())
    return args


def WIN(func, restype, *argtypes) -> None:
    func.restype = restype
    func.argtypes = argtypes
    if issubclass(restype, NTSTATUS):
        func.errcheck = _check_status
    elif issubclass(restype, BOOL):
        func.errcheck = _check_bool


# https://msdn.microsoft.com/en-us/library/ms683179
WIN(kernel32.GetCurrentProcess, HANDLE)

# https://msdn.microsoft.com/en-us/library/ms724251
WIN(kernel32.DuplicateHandle, BOOL,
    HANDLE,  # _In_  hSourceProcessHandle
    HANDLE,  # _In_  hSourceHandle
    HANDLE,  # _In_  hTargetProcessHandle
    PHANDLE,  # _Out_ lpTargetHandle
    DWORD,  # _In_  dwDesiredAccess
    BOOL,  # _In_  bInheritHandle
    DWORD)  # _In_  dwOptions

# https://msdn.microsoft.com/en-us/library/ms724295
WIN(kernel32.GetComputerNameW, BOOL,
    LPWSTR,  # _Out_   lpBuffer
    LPDWORD)  # _Inout_ lpnSize

# https://msdn.microsoft.com/en-us/library/aa379295
WIN(advapi32.OpenProcessToken, BOOL,
    HANDLE,  # _In_  ProcessHandle
    DWORD,  # _In_  DesiredAccess
    PHANDLE)  # _Out_ TokenHandle

# https://msdn.microsoft.com/en-us/library/aa446617
WIN(advapi32.DuplicateTokenEx, BOOL,
    HANDLE,  # _In_     hExistingToken
    DWORD,  # _In_     dwDesiredAccess
    LPSECURITY_ATTRIBUTES,  # _In_opt_ lpTokenAttributes
    SECURITY_IMPERSONATION_LEVEL,  # _In_     ImpersonationLevel
    TOKEN_TYPE,  # _In_     TokenType
    PHANDLE)  # _Out_    phNewToken

# https://msdn.microsoft.com/en-us/library/ff566415
WIN(ntdll.NtAllocateLocallyUniqueId, NTSTATUS,
    PLUID)  # _Out_ LUID

# https://msdn.microsoft.com/en-us/library/aa378279
WIN(secur32.LsaFreeReturnBuffer, NTSTATUS,
    LPVOID, )  # _In_ Buffer

# https://msdn.microsoft.com/en-us/library/aa378265
WIN(secur32.LsaConnectUntrusted, NTSTATUS,
    PHANDLE, )  # _Out_ LsaHandle

#https://msdn.microsoft.com/en-us/library/aa378318
WIN(secur32.LsaRegisterLogonProcess, NTSTATUS,
    PSTRING,  # _In_  LogonProcessName
    PHANDLE,  # _Out_ LsaHandle
    PLSA_OPERATIONAL_MODE)  # _Out_ SecurityMode

# https://msdn.microsoft.com/en-us/library/aa378269
WIN(secur32.LsaDeregisterLogonProcess, NTSTATUS,
    HANDLE)  # _In_ LsaHandle

# https://msdn.microsoft.com/en-us/library/aa378297
WIN(secur32.LsaLookupAuthenticationPackage, NTSTATUS,
    HANDLE,  # _In_  LsaHandle
    PSTRING,  # _In_  PackageName
    PULONG)  # _Out_ AuthenticationPackage

# https://msdn.microsoft.com/en-us/library/aa378292
WIN(secur32.LsaLogonUser, NTSTATUS,
    HANDLE,  # _In_     LsaHandle
    PSTRING,  # _In_     OriginName
    SECURITY_LOGON_TYPE,  # _In_     LogonType
    ULONG,  # _In_     AuthenticationPackage
    LPVOID,  # _In_     AuthenticationInformation
    ULONG,  # _In_     AuthenticationInformationLength
    PTOKEN_GROUPS,  # _In_opt_ LocalGroups
    PTOKEN_SOURCE,  # _In_     SourceContext
    PLPVOID,  # _Out_    ProfileBuffer
    PULONG,  # _Out_    ProfileBufferLength
    PLUID,  # _Out_    LogonId
    PHANDLE,  # _Out_    Token
    PQUOTA_LIMITS,  # _Out_    Quotas
    PNTSTATUS)  # _Out_    SubStatus


def duplicate_token(source_token=None, access=TOKEN_ALL_ACCESS,
                    impersonation_level=SecurityImpersonation,
                    token_type=TokenPrimary, attributes=None) -> MANAGED_HANDLE:
    close_source = False
    if source_token is None:
        close_source = True
        source_token = MANAGED_HANDLE()
        advapi32.OpenProcessToken(kernel32.GetCurrentProcess(), TOKEN_ALL_ACCESS, byref(source_token))
    token = MANAGED_HANDLE()
    try:
        advapi32.DuplicateTokenEx(source_token, access, attributes, impersonation_level, token_type, byref(token))
    finally:
        if close_source:
            source_token.close()
    return token


def lsa_connect_untrusted() -> int:
    handle = HANDLE()
    secur32.LsaConnectUntrusted(byref(handle))
    return handle.value


def lsa_register_logon_process(logon_process_name) -> int:
    if not isinstance(logon_process_name, bytes):
        logon_process_name = logon_process_name.encode('mbcs')
    logon_process_name = logon_process_name[:127]
    buf = create_string_buffer(logon_process_name, 128)
    name = STRING(len(logon_process_name), len(buf), buf)
    handle = HANDLE()
    mode = LSA_OPERATIONAL_MODE()
    secur32.LsaRegisterLogonProcess(byref(name), byref(handle), byref(mode))
    return handle.value


def lsa_lookup_authentication_package(lsa_handle, package_name):
    if not isinstance(package_name, bytes):
        package_name = package_name.encode('mbcs')
    package_name = package_name[:127]
    buf = create_string_buffer(package_name)
    name = STRING(len(package_name), len(buf), buf)
    package = ULONG()
    secur32.LsaLookupAuthenticationPackage(lsa_handle, byref(name), byref(package))
    return package.value


# Low-level LSA logon

LOGONINFO = namedtuple('LOGONINFO', ('Token', 'LogonId', 'Profile', 'Quotas'))


def lsa_logon_user(auth_info, local_groups=None, origin_name=py_origin_name,
                   source_context=None, auth_package=None, logon_type=None,
                   lsa_handle=None) -> LOGONINFO:
    log("lsa_logon_user%s", (auth_info, local_groups, origin_name,
                             source_context, auth_package, logon_type,
                             lsa_handle))
    if local_groups is None:
        plocal_groups = PTOKEN_GROUPS()
    else:
        plocal_groups = byref(local_groups)
    if source_context is None:
        source_context = py_source_context
    if not isinstance(origin_name, bytes):
        origin_name = origin_name.encode('mbcs')
    buf = create_string_buffer(origin_name)
    origin_name = STRING(len(origin_name), len(buf), buf)
    if auth_package is None:
        if isinstance(auth_info, MSV1_0_S4U_LOGON):
            auth_package = NEGOTIATE_PACKAGE_NAME
        elif isinstance(auth_info, KERB_S4U_LOGON):
            auth_package = MICROSOFT_KERBEROS_NAME
        else:
            auth_package = MSV1_0_PACKAGE_NAME
    if logon_type is None:
        if isinstance(auth_info, S4ULogon):
            logon_type = Batch
        else:
            logon_type = Interactive
    profile_buffer = LPVOID()
    profile_buffer_length = ULONG()
    profile = None
    logonid = LUID()
    htoken = MANAGED_HANDLE()
    quotas = QUOTA_LIMITS()
    substatus = NTSTATUS()
    deregister = False
    if lsa_handle is None:
        lsa_handle = lsa_connect_untrusted()
        deregister = True
    try:
        if isinstance(auth_package, (str, bytes)):
            auth_package = lsa_lookup_authentication_package(lsa_handle, auth_package)
        try:
            args = (
                lsa_handle, byref(origin_name),
                logon_type, auth_package, byref(auth_info),
                sizeof(auth_info), plocal_groups,
                byref(source_context), byref(profile_buffer),
                byref(profile_buffer_length), byref(logonid),
                byref(htoken), byref(quotas),
                byref(substatus),
            )
            log("LsaLogonUser%s", args)
            secur32.LsaLogonUser(*args)
        except OSError:
            if substatus.value:
                raise WinError(substatus.to_error()) from None
            raise
        finally:
            if profile_buffer:
                address = profile_buffer.value
                buftype = PROFILE_BUFFER_TYPE.from_address(address).value
                if buftype == MsV1_0InteractiveLogon:
                    profile = MSV1_0_INTERACTIVE_PROFILE.from_address_copy(address, profile_buffer_length.value)
                secur32.LsaFreeReturnBuffer(address)
    finally:
        if deregister:
            secur32.LsaDeregisterLogonProcess(lsa_handle)
    return LOGONINFO(htoken, logonid, profile, quotas)


# High-level LSA logons

def logon_msv1(name: str, password: str, domain=None, local_groups=None,
               origin_name=py_origin_name, source_context=None) -> LOGONINFO:
    return lsa_logon_user(MSV1_0_INTERACTIVE_LOGON(name, password, domain),
                          local_groups, origin_name, source_context)


def logon_msv1_s4u(name: str, local_groups=None,
                   origin_name=py_origin_name,
                   source_context=None) -> LOGONINFO:
    domain = create_unicode_buffer(MAX_COMPUTER_NAME_LENGTH + 1)
    length = DWORD(len(domain))
    kernel32.GetComputerNameW(domain, byref(length))
    return lsa_logon_user(MSV1_0_S4U_LOGON(name, domain.value),
                          local_groups, origin_name, source_context)


def logon_kerb_s4u(name: str, realm=None, local_groups=None,
                   origin_name=py_origin_name,
                   source_context=None,
                   logon_process_name=py_logon_process_name) -> LOGONINFO:
    lsa_handle = lsa_register_logon_process(logon_process_name)
    try:
        return lsa_logon_user(KERB_S4U_LOGON(name, realm),
                              local_groups, origin_name, source_context,
                              lsa_handle=lsa_handle)
    finally:
        secur32.LsaDeregisterLogonProcess(lsa_handle)


def main() -> int:
    from xpra.platform import program_context
    from xpra.log import enable_color, consume_verbose_argv
    with program_context("LSA-Logon-Test", "LSA Logon Test"):
        enable_color()
        consume_verbose_argv(sys.argv, "win32")
        if len(sys.argv) not in (2, 3):
            log.warn("invalid number of arguments")
            log.warn("usage: %s [--verbose] username [password]", sys.argv[0])
            return 1
        username = sys.argv[1]
        password = "" if len(sys.argv) < 3 else sys.argv[2]
        try:
            if password:
                logon_msv1(username, password)
            else:
                logon_msv1_s4u(username)
            return 0
        except Exception as e:
            log.error("Logon failed: %s", e)
            return 1


if __name__ == "__main__":
    sys.exit(main())



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any
from collections.abc import Sequence

from xpra.util.str_fn import csv, Ellipsizer, bytestostr
from xpra.os_util import gi_import
from xpra.dbus.helper import native_to_dbus
from xpra.notifications.notifier_base import NotifierBase, log, NID
import dbus
from dbus.mainloop.glib import DBusGMainLoop, threads_init
from dbus.exceptions import DBusException

threads_init()
DBusGMainLoop(set_as_default=True)

NOTIFICATION_APP_NAME = os.environ.get("XPRA_NOTIFICATION_APP_NAME", "%s (via Xpra)")
FD_NOTIFICATIONS = 'org.freedesktop.Notifications'


def DBUS_Notifier_factory(*args):
    try:
        return DBUS_Notifier(*args)
    except Exception as e:
        log.warn("Warning: failed to instantiate the dbus notification handler")
        if str(e).startswith("org.freedesktop.DBus.Error.ServiceUnknown:"):
            log.warn(" you may need to start a notification service for 'org.freedesktop.Notifications'")
        else:
            log.warn(" %s", e)
        log.warn(" disable notifications to avoid this warning")
        return None


class DBUS_Notifier(NotifierBase):

    def __init__(self, *args):
        super().__init__(*args)
        self.app_name_format = NOTIFICATION_APP_NAME
        self.last_notification: Sequence[Any] = ()
        self.actual_notification_id: dict[int, int] = {}
        self.dbusnotify = None
        self.setup_dbusnotify()
        self.handles_actions = True
        self.may_retry = True

    def setup_dbusnotify(self) -> None:
        self.dbus_session = dbus.SessionBus()
        self.org_fd_notifications = self.dbus_session.get_object(FD_NOTIFICATIONS, '/org/freedesktop/Notifications')
        self.org_fd_notifications.connect_to_signal("NotificationClosed", self.NotificationClosed)
        self.org_fd_notifications.connect_to_signal("ActionInvoked", self.ActionInvoked)

        # connect_to_signal("HelloSignal", hello_signal_handler, dbus_interface="com.example.TestService", arg0="Hello")
        self.dbusnotify = dbus.Interface(self.org_fd_notifications, FD_NOTIFICATIONS)
        log("using dbusnotify: %s(%s)", type(self.dbusnotify), FD_NOTIFICATIONS)
        caps = tuple(str(x) for x in self.dbusnotify.GetCapabilities())
        log("capabilities=%s", csv(caps))
        self.handles_actions = "actions" in caps
        log("dbus.get_default_main_loop()=%s", dbus.get_default_main_loop())

    def cleanup(self) -> None:
        nids = list(self.actual_notification_id.items())
        self.actual_notification_id = {}
        for nid, actual_id in nids:
            self.do_close(nid, actual_id)
        super().cleanup()
        self.dbusnotify = None

    def show_notify(self, dbus_id, tray, nid: NID,
                    app_name: str, replaces_nid: NID, app_icon,
                    summary: str, body: str, actions, hints, timeout: int, icon) -> None:
        if not self.dbus_check(dbus_id):
            return
        if not self.dbusnotify:
            return
        self.may_retry = True
        with log.trap_error("Error: dbus notify failed"):
            icon_string = self.get_icon_string(nid, app_icon, icon)
            log("get_icon_string%s=%s", (nid, app_icon, Ellipsizer(icon)), icon_string)
            if app_name == "Xpra":
                # don't show "Xpra (via Xpra)"
                app_str = "Xpra"
            else:
                try:
                    app_str = self.app_name_format % app_name
                except TypeError:
                    app_str = app_name or "Xpra"
            self.last_notification = (
                dbus_id, tray, nid, app_name, replaces_nid,
                app_icon, summary, body, actions, hints, timeout, icon,
            )

            def NotifyReply(notification_id):
                log("NotifyReply(%s) for nid=%i", notification_id, nid)
                self.actual_notification_id[int(nid)] = int(notification_id)

            dbus_hints = self.parse_hints(hints)
            log("calling %s%s", self.dbusnotify.Notify,
                (app_str, 0, icon_string, summary, body, actions, dbus_hints, timeout))
            self.dbusnotify.Notify(app_str, 0, icon_string, summary, body, actions, dbus_hints, timeout,
                                   reply_handler=NotifyReply,
                                   error_handler=self.NotifyError)

    def _find_nid(self, actual_id) -> int:
        aid = int(actual_id)
        for k, v in self.actual_notification_id.items():
            if v == aid:
                return k
        return 0

    def noparse_hints(self, h) -> dict:
        return h

    def parse_hints(self, h) -> dbus.types.Dictionary:
        hints = {}
        for x in ("action-icons", "category", "desktop-entry", "resident", "transient", "x", "y", "urgency"):
            v = h.get(x)
            if v is not None:
                hints[native_to_dbus(x)] = native_to_dbus(v)
        image_data = h.get("image-data")
        if image_data and bytestostr(image_data[0]) == "png":
            try:
                from xpra.codecs.pillow.decoder import open_only  # pylint: disable=import-outside-toplevel
                img_data = image_data[3]
                img = open_only(img_data, ("png",))
                w, h = img.size
                channels = len(img.mode)
                rowstride = w * channels
                has_alpha = img.mode == "RGBA"
                pixel_data = bytearray(img.tobytes("raw", img.mode))
                args = w, h, rowstride, has_alpha, 8, channels, pixel_data
                hints["image-data"] = tuple(native_to_dbus(x) for x in args)
            except Exception as e:
                log("parse_hints(%s) error on image-data=%s", h, image_data, exc_info=True)
                log.error("Error parsing notification image:")
                log.estr(e)
        log("parse_hints(%s)=%s", h, hints)
        return dbus.types.Dictionary(hints, signature="sv")

    def NotificationClosed(self, actual_id: dbus.UInt32, reason) -> None:
        nid = self._find_nid(actual_id)
        reason_str = {
            1: "expired",
            2: "dismissed by the user",
            3: "closed by a call to CloseNotification",
            4: "Undefined/reserved reasons",
        }.get(int(reason), str(reason))
        log("NotificationClosed(%s, %s) nid=%s, reason=%s", actual_id, reason, nid, reason_str)
        if nid:
            self.actual_notification_id.pop(nid, None)
            self.clean_notification(nid)
            if self.closed_cb:
                self.closed_cb(nid, int(reason), reason_str)

    def ActionInvoked(self, actual_id: dbus.UInt32, action) -> None:
        nid = self._find_nid(actual_id)
        log("ActionInvoked(%s, %s) nid=%s", actual_id, action, nid)
        if nid and self.action_cb:
            self.action_cb(nid, str(action))

    def NotifyError(self, dbus_error, *_args) -> bool:
        if not self.dbusnotify:
            return False
        try:
            if isinstance(dbus_error, DBusException):
                message = dbus_error.get_dbus_message()
                dbus_error_name = dbus_error.get_dbus_name()
                if dbus_error_name != "org.freedesktop.DBus.Error.ServiceUnknown":
                    log.error("unhandled dbus exception: %s, %s", message, dbus_error_name)
                    return False

                if not self.may_retry:
                    log.error("Error: cannot send notification via dbus,")
                    log.error(" check that you notification service is operating properly")
                    return False
                self.may_retry = False

                log.info("trying to re-connect to the notification service")
                # try to connect to the notification again (just once):
                self.setup_dbusnotify()
                # and retry:
                self.show_notify(*self.last_notification)
        except Exception:
            log("cannot filter error", exc_info=True)
        log.error("Error processing notification:")
        log.estr(dbus_error)
        return False

    def close_notify(self, nid: NID) -> None:
        actual_id = self.actual_notification_id.get(int(nid))
        if actual_id is None:
            log("close_notify(%i) actual notification not found, already closed?", nid)
            return
        log("close_notify(%i) actual id=%s", nid, actual_id)
        self.do_close(nid, actual_id)

    def do_close(self, _nid: NID, actual_id: int) -> None:
        log("do_close_notify(%i)", actual_id)
        if not self.dbusnotify:
            return

        def CloseNotificationReply():
            self.actual_notification_id.pop(actual_id, None)

        def CloseNotificationError(dbus_error, *_args):
            log.warn("Error: error closing notification:")
            log.warn(" %s", dbus_error)

        self.dbusnotify.CloseNotification(actual_id,
                                          reply_handler=CloseNotificationReply,
                                          error_handler=CloseNotificationError)


def main():
    # pylint: disable=import-outside-toplevel
    Gtk = gi_import("Gtk")
    GLib = gi_import("GLib")

    def show():
        n = DBUS_Notifier_factory()
        # actions = ["0", "Hello", "1", "Bye"]
        actions = []
        n.show_notify("", None, 0, "Test", 0, "",
                      "Summary", "Body line1\nline2...",
                      actions, {}, 0, "")
        return False

    GLib.idle_add(show)
    GLib.timeout_add(20000, Gtk.main_quit)
    Gtk.main()


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.os_util import gi_import
from xpra.common import noop
from xpra.clipboard.core import ClipboardProtocolHelperCore, ClipboardProxyCore
from xpra.util.str_fn import Ellipsizer, repr_ellipsized
from xpra.util.env import envint
from xpra.log import Logger
from xpra.platform.features import CLIPBOARD_GREEDY

GLib = gi_import("GLib")

log = Logger("clipboard")


def env_timeout(name, default: int, min_time=0, max_time=5000) -> int:
    env_name = f"XPRA_CLIPBOARD_{name}_TIMEOUT"
    value = envint(env_name, default)
    if not min_time < value <= max_time:
        log.warn(f"Warning: invalid value for {env_name!r}")
        log.warn(f" valid range is from {min_time} to {max_time}")
        value = max(min_time, min(max_time, value))
    return value


CONVERT_TIMEOUT = env_timeout("CONVERT", 100)
REMOTE_TIMEOUT = env_timeout("REMOTE", 2500)


class ClipboardTimeoutHelper(ClipboardProtocolHelperCore):
    # a clipboard superclass that handles timeouts

    def __init__(self, send_packet_cb, progress_cb=noop, **kwargs):
        super().__init__(send_packet_cb, progress_cb, **kwargs)
        self._clipboard_outstanding_requests: dict[int, tuple[int, str, str]] = {}

    def cleanup(self) -> None:
        # reply to outstanding requests with "no data":
        for request_id in tuple(self._clipboard_outstanding_requests.keys()):
            self._clipboard_got_contents(request_id)
        self._clipboard_outstanding_requests = {}
        super().cleanup()

    def make_proxy(self, selection: str):
        raise NotImplementedError()

    def _get_proxy(self, selection: str) -> ClipboardProxyCore | None:
        proxy = self._clipboard_proxies.get(selection)
        if not proxy:
            log.warn("Warning: no clipboard proxy for '%s'", selection)
            return None
        return proxy

    def set_want_targets_client(self, want_targets: bool) -> None:
        super().set_want_targets_client(want_targets)
        # pass it on to the ClipboardProxy instances:
        for proxy in self._clipboard_proxies.values():
            proxy.set_want_targets(want_targets)

    ############################################################################
    # network methods for communicating with the remote clipboard:
    ############################################################################
    def _send_clipboard_token_handler(self, proxy: ClipboardProxyCore, packet_data=()):
        if log.is_debug_enabled():
            log("_send_clipboard_token_handler(%s, %s)", proxy, repr_ellipsized(packet_data))
        remote = self.local_to_remote(proxy._selection)
        packet: list[Any] = ["clipboard-token", remote]
        if packet_data:
            # append 'TARGETS' unchanged:
            packet.append(packet_data[0])
            # if present, the next element is the target data,
            # which we have to convert to wire format:
            if len(packet_data) >= 2:
                target, dtype, dformat, data = packet_data[1]
                wire_encoding, wire_data = self._munge_raw_selection_to_wire(target, dtype, dformat, data)
                if wire_encoding:
                    wire_data = self._may_compress(dtype, dformat, wire_data)
                    if wire_data:
                        packet += [target, dtype, dformat, wire_encoding, wire_data]
                        claim = proxy._can_send
                        packet += [claim, CLIPBOARD_GREEDY]
        log("send_clipboard_token_handler %s to %s", proxy._selection, remote)
        self.send(*packet)

    def _send_clipboard_request_handler(self, proxy: ClipboardProxyCore, selection: str, target: str):
        log("send_clipboard_request_handler%s", (proxy, selection, target))
        request_id = self._clipboard_request_counter
        self._clipboard_request_counter += 1
        remote = self.local_to_remote(selection)
        log("send_clipboard_request %s to %s, id=%s", selection, remote, request_id)
        timer = GLib.timeout_add(REMOTE_TIMEOUT, self.timeout_request, request_id)
        self._clipboard_outstanding_requests[request_id] = (timer, selection, target)
        self.progress()
        self.send("clipboard-request", request_id, remote, target)

    def timeout_request(self, request_id: int) -> None:
        try:
            selection, target = self._clipboard_outstanding_requests.pop(request_id)[1:]
        except KeyError:
            log.warn("Warning: clipboard request id %i not found", request_id)
            return
        finally:
            self.progress()
        log.warn("Warning: remote clipboard request timed out")
        log.warn(" request id %i, selection=%s, target=%s", request_id, selection, target)
        proxy = self._get_proxy(selection)
        if proxy:
            proxy.got_contents(target)

    def _clipboard_got_contents(self, request_id: int, dtype: str = "", dformat: int = 0, data=None) -> None:
        try:
            timer, selection, target = self._clipboard_outstanding_requests.pop(request_id)
        except KeyError:
            log.warn("Warning: request id %i not found", request_id)
            log.warn(" already timed out or duplicate reply")
            return
        finally:
            self.progress()
        GLib.source_remove(timer)
        proxy = self._get_proxy(selection)
        log("clipboard got contents%s: proxy=%s for selection=%s",
            (request_id, dtype, dformat, Ellipsizer(data)), proxy, selection)
        if data and isinstance(data, memoryview):
            data = bytes(data)
        if proxy:
            proxy.got_contents(target, dtype, dformat, data)

    def client_reset(self) -> None:
        super().client_reset()
        # timeout all pending requests
        cor = self._clipboard_outstanding_requests
        if cor:
            log.info("cancelling %i clipboard requests", len(cor))
            self._clipboard_outstanding_requests = {}
            for request_id in cor:
                self._clipboard_got_contents(request_id)
